# -*- encoding : utf-8 -*-
########################################################
## Thoughts from reading the ISO 32000-1:2008
## this file is part of the CombinePDF library and the code
## is subject to the same license.
########################################################






module CombinePDF

	#:nodoc: all



	module Fonts

		protected

		# the internal class for the Fonts model
		#
		# this is an internal class, used by PDFWriter and PDF. you don't normally need to use this.
		class Font < Hash
			# set/get the name of the font
			attr_accessor :name
			# get the metrics dictionary for the font
			attr_reader :metrics
			# set the metrics dictionary of the font, making sure the :missing value is set.
			def metrics= (new_metrics)
				@metrics = new_metrics
				@metrics[:missing] = @metrics.first[1] unless @metrics[:missing]
			end
			# internelized a new Font object, setting it's name, it's metrics Hash and the object Hash.
			def initialize(name = nil, font_metrics = {32=>{:wx=>250, :boundingbox=>[0, 0, 0, 0]}}, object = nil)
				self.name = name
				self.metrics = font_metrics
				self[:is_reference_only] = true
				self[:referenced_object] = object
			end

			# This function calculates the dimensions of a string in a PDF.
			#
			# UNICODE SUPPORT IS FONT DEPENDENT!
			#
			# text:: String containing the text for which the demantion box will be calculated.
			# size:: the size of the text, as it will be applied in the PDF.
			def dimensions_of(text, size = 1000)
				metrics_array = []
				# the following is only good for latin text - unicode support is missing!!!!
				text.each_char do |c|
					metrics_array << (self.metrics[c.ord] or self.metrics[:missing])
				end
				height = metrics_array.map {|m| m ? m[:boundingbox][3] : 0} .max
				height = height - (metrics_array.map {|m| m ? m[:boundingbox][1] : 0} ).min
				width = 0.0
				metrics_array.each do |m|
					width += (m[:wx] or m[:wy])
				end
				return [height.to_f/1000*size, width.to_f/1000*size] if metrics_array[0][:wy]
				[width.to_f/1000*size, height.to_f/1000*size]
			end
		end
	end
end



module CombinePDF

	#:nodoc: all

	# this is an internal module used to store the fonts used by PDFWriter. You dont need to use this module.
	#
	# In the future, this model will help to add unicode supports and fonts support,
	# so that the page numbering and textbox features could be used with international fonts and types.
	#
	# this model will be used in the future for the add_font method which will be defined in CombinePDF or PDFWriter.
	module Fonts
		module_function

		# gets a font from the fonts library
		def get_font(name = :Helvetica)
			initiate_library
			FONTS_LIBRARY[name]
		end

		# adds a correctly formatted font object to the font library.
		# font_name:: a Symbol with the name of the font. if the fonts exists, it will be overwritten!
		# font_metrics:: a Hash of ont metrics, of the format i => {wx: char_width, boundingbox: [left_x, buttom_y, right_x, top_y]} where i == character code (i.e. 32 for space). The Hash should contain a special value :missing for the metrics of missing characters. an optional :wy will be supported in the future, for up to down fonts.
		# font_pdf_object:: a Hash in the internal format recognized by CombinePDF, that represents the font object.
		def register_font(font_name, font_metrics, font_pdf_object)
			new_font = Font.new
			new_font.name = font_name
			new_font.metrics = font_metrics
			new_font[:is_reference_only] = true
			new_font[:referenced_object] = font_pdf_object
			FONTS_LIBRARY[new_font.name] = new_font
		end

		# This function calculates the dimensions of a string in a PDF.
		#
		# UNICODE SUPPORT IS MISSING!
		#
		# text:: String containing the text for which the demantion box will be calculated.
		# font:: the font name, from the 14 fonts possible. @see font
		# size:: the size of the text, as it will be applied in the PDF.
		def dimensions_of(text, font_name, size = 1000)
			get_font(font_name).dimensions_of text, size
		end

		# this function registers the 14 standard fonts to the library.
		#
		# it will be called when the module first requests a font from the library.
		#
		# if the 14 standard fonts are already registered, it will simply return false.
		def initiate_library
			# do nothing if the library is already initiated
			return false if FONTS_LIBRARY.has_key? :Helvetica
			# font metrics objects to be used
			times_metrics = {nil=>{:wx=>250, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>333, :boundingbox=>[130, -9, 238, 676]}, 34=>{:wx=>408, :boundingbox=>[77, 431, 331, 676]}, 35=>{:wx=>500, :boundingbox=>[5, 0, 496, 662]}, 36=>{:wx=>500, :boundingbox=>[44, -87, 457, 727]}, 37=>{:wx=>833, :boundingbox=>[61, -13, 772, 676]}, 38=>{:wx=>778, :boundingbox=>[42, -13, 750, 676]}, 39=>{:wx=>333, :boundingbox=>[79, 433, 218, 676]}, 40=>{:wx=>333, :boundingbox=>[48, -177, 304, 676]}, 41=>{:wx=>333, :boundingbox=>[29, -177, 285, 676]}, 42=>{:wx=>500, :boundingbox=>[69, 265, 432, 676]}, 43=>{:wx=>564, :boundingbox=>[30, 0, 534, 506]}, 44=>{:wx=>250, :boundingbox=>[56, -141, 195, 102]}, 45=>{:wx=>333, :boundingbox=>[39, 194, 285, 257]}, 46=>{:wx=>250, :boundingbox=>[70, -11, 181, 100]}, 47=>{:wx=>278, :boundingbox=>[-9, -14, 287, 676]}, 48=>{:wx=>500, :boundingbox=>[24, -14, 476, 676]}, 49=>{:wx=>500, :boundingbox=>[111, 0, 394, 676]}, 50=>{:wx=>500, :boundingbox=>[30, 0, 475, 676]}, 51=>{:wx=>500, :boundingbox=>[43, -14, 431, 676]}, 52=>{:wx=>500, :boundingbox=>[12, 0, 472, 676]}, 53=>{:wx=>500, :boundingbox=>[32, -14, 438, 688]}, 54=>{:wx=>500, :boundingbox=>[34, -14, 468, 684]}, 55=>{:wx=>500, :boundingbox=>[20, -8, 449, 662]}, 56=>{:wx=>500, :boundingbox=>[56, -14, 445, 676]}, 57=>{:wx=>500, :boundingbox=>[30, -22, 459, 676]}, 58=>{:wx=>278, :boundingbox=>[81, -11, 192, 459]}, 59=>{:wx=>278, :boundingbox=>[80, -141, 219, 459]}, 60=>{:wx=>564, :boundingbox=>[28, -8, 536, 514]}, 61=>{:wx=>564, :boundingbox=>[30, 120, 534, 386]}, 62=>{:wx=>564, :boundingbox=>[28, -8, 536, 514]}, 63=>{:wx=>444, :boundingbox=>[68, -8, 414, 676]}, 64=>{:wx=>921, :boundingbox=>[116, -14, 809, 676]}, 65=>{:wx=>722, :boundingbox=>[15, 0, 706, 674]}, 66=>{:wx=>667, :boundingbox=>[17, 0, 593, 662]}, 67=>{:wx=>667, :boundingbox=>[28, -14, 633, 676]}, 68=>{:wx=>722, :boundingbox=>[16, 0, 685, 662]}, 69=>{:wx=>611, :boundingbox=>[12, 0, 597, 662]}, 70=>{:wx=>556, :boundingbox=>[12, 0, 546, 662]}, 71=>{:wx=>722, :boundingbox=>[32, -14, 709, 676]}, 72=>{:wx=>722, :boundingbox=>[19, 0, 702, 662]}, 73=>{:wx=>333, :boundingbox=>[18, 0, 315, 662]}, 74=>{:wx=>389, :boundingbox=>[10, -14, 370, 662]}, 75=>{:wx=>722, :boundingbox=>[34, 0, 723, 662]}, 76=>{:wx=>611, :boundingbox=>[12, 0, 598, 662]}, 77=>{:wx=>889, :boundingbox=>[12, 0, 863, 662]}, 78=>{:wx=>722, :boundingbox=>[12, -11, 707, 662]}, 79=>{:wx=>722, :boundingbox=>[34, -14, 688, 676]}, 80=>{:wx=>556, :boundingbox=>[16, 0, 542, 662]}, 81=>{:wx=>722, :boundingbox=>[34, -178, 701, 676]}, 82=>{:wx=>667, :boundingbox=>[17, 0, 659, 662]}, 83=>{:wx=>556, :boundingbox=>[42, -14, 491, 676]}, 84=>{:wx=>611, :boundingbox=>[17, 0, 593, 662]}, 85=>{:wx=>722, :boundingbox=>[14, -14, 705, 662]}, 86=>{:wx=>722, :boundingbox=>[16, -11, 697, 662]}, 87=>{:wx=>944, :boundingbox=>[5, -11, 932, 662]}, 88=>{:wx=>722, :boundingbox=>[10, 0, 704, 662]}, 89=>{:wx=>722, :boundingbox=>[22, 0, 703, 662]}, 90=>{:wx=>611, :boundingbox=>[9, 0, 597, 662]}, 91=>{:wx=>333, :boundingbox=>[88, -156, 299, 662]}, 92=>{:wx=>278, :boundingbox=>[-9, -14, 287, 676]}, 93=>{:wx=>333, :boundingbox=>[34, -156, 245, 662]}, 94=>{:wx=>469, :boundingbox=>[24, 297, 446, 662]}, 95=>{:wx=>500, :boundingbox=>[0, -125, 500, -75]}, 96=>{:wx=>333, :boundingbox=>[115, 433, 254, 676]}, 97=>{:wx=>444, :boundingbox=>[37, -10, 442, 460]}, 98=>{:wx=>500, :boundingbox=>[3, -10, 468, 683]}, 99=>{:wx=>444, :boundingbox=>[25, -10, 412, 460]}, 100=>{:wx=>500, :boundingbox=>[27, -10, 491, 683]}, 101=>{:wx=>444, :boundingbox=>[25, -10, 424, 460]}, 102=>{:wx=>333, :boundingbox=>[20, 0, 383, 683]}, 103=>{:wx=>500, :boundingbox=>[28, -218, 470, 460]}, 104=>{:wx=>500, :boundingbox=>[9, 0, 487, 683]}, 105=>{:wx=>278, :boundingbox=>[16, 0, 253, 683]}, 106=>{:wx=>278, :boundingbox=>[-70, -218, 194, 683]}, 107=>{:wx=>500, :boundingbox=>[7, 0, 505, 683]}, 108=>{:wx=>278, :boundingbox=>[19, 0, 257, 683]}, 109=>{:wx=>778, :boundingbox=>[16, 0, 775, 460]}, 110=>{:wx=>500, :boundingbox=>[16, 0, 485, 460]}, 111=>{:wx=>500, :boundingbox=>[29, -10, 470, 460]}, 112=>{:wx=>500, :boundingbox=>[5, -217, 470, 460]}, 113=>{:wx=>500, :boundingbox=>[24, -217, 488, 460]}, 114=>{:wx=>333, :boundingbox=>[5, 0, 335, 460]}, 115=>{:wx=>389, :boundingbox=>[51, -10, 348, 460]}, 116=>{:wx=>278, :boundingbox=>[13, -10, 279, 579]}, 117=>{:wx=>500, :boundingbox=>[9, -10, 479, 450]}, 118=>{:wx=>500, :boundingbox=>[19, -14, 477, 450]}, 119=>{:wx=>722, :boundingbox=>[21, -14, 694, 450]}, 120=>{:wx=>500, :boundingbox=>[17, 0, 479, 450]}, 121=>{:wx=>500, :boundingbox=>[14, -218, 475, 450]}, 122=>{:wx=>444, :boundingbox=>[27, 0, 418, 450]}, 123=>{:wx=>480, :boundingbox=>[100, -181, 350, 680]}, 124=>{:wx=>200, :boundingbox=>[67, -218, 133, 782]}, 125=>{:wx=>480, :boundingbox=>[130, -181, 380, 680]}, 126=>{:wx=>541, :boundingbox=>[40, 183, 502, 323]}, 161=>{:wx=>333, :boundingbox=>[97, -218, 205, 467]}, 162=>{:wx=>500, :boundingbox=>[53, -138, 448, 579]}, 163=>{:wx=>500, :boundingbox=>[12, -8, 490, 676]}, 164=>{:wx=>167, :boundingbox=>[-168, -14, 331, 676]}, 165=>{:wx=>500, :boundingbox=>[-53, 0, 512, 662]}, 166=>{:wx=>500, :boundingbox=>[7, -189, 490, 676]}, 167=>{:wx=>500, :boundingbox=>[70, -148, 426, 676]}, 168=>{:wx=>500, :boundingbox=>[-22, 58, 522, 602]}, 169=>{:wx=>180, :boundingbox=>[48, 431, 133, 676]}, 170=>{:wx=>444, :boundingbox=>[43, 433, 414, 676]}, 171=>{:wx=>500, :boundingbox=>[42, 33, 456, 416]}, 172=>{:wx=>333, :boundingbox=>[63, 33, 285, 416]}, 173=>{:wx=>333, :boundingbox=>[48, 33, 270, 416]}, 174=>{:wx=>556, :boundingbox=>[31, 0, 521, 683]}, 175=>{:wx=>556, :boundingbox=>[32, 0, 521, 683]}, 177=>{:wx=>500, :boundingbox=>[0, 201, 500, 250]}, 178=>{:wx=>500, :boundingbox=>[59, -149, 442, 676]}, 179=>{:wx=>500, :boundingbox=>[58, -153, 442, 676]}, 180=>{:wx=>250, :boundingbox=>[70, 199, 181, 310]}, 182=>{:wx=>453, :boundingbox=>[-22, -154, 450, 662]}, 183=>{:wx=>350, :boundingbox=>[40, 196, 310, 466]}, 184=>{:wx=>333, :boundingbox=>[79, -141, 218, 102]}, 185=>{:wx=>444, :boundingbox=>[45, -141, 416, 102]}, 186=>{:wx=>444, :boundingbox=>[30, 433, 401, 676]}, 187=>{:wx=>500, :boundingbox=>[44, 33, 458, 416]}, 188=>{:wx=>1000, :boundingbox=>[111, -11, 888, 100]}, 189=>{:wx=>1000, :boundingbox=>[7, -19, 994, 706]}, 191=>{:wx=>444, :boundingbox=>[30, -218, 376, 466]}, 193=>{:wx=>333, :boundingbox=>[19, 507, 242, 678]}, 194=>{:wx=>333, :boundingbox=>[93, 507, 317, 678]}, 195=>{:wx=>333, :boundingbox=>[11, 507, 322, 674]}, 196=>{:wx=>333, :boundingbox=>[1, 532, 331, 638]}, 197=>{:wx=>333, :boundingbox=>[11, 547, 322, 601]}, 198=>{:wx=>333, :boundingbox=>[26, 507, 307, 664]}, 199=>{:wx=>333, :boundingbox=>[118, 581, 216, 681]}, 200=>{:wx=>333, :boundingbox=>[18, 581, 315, 681]}, 202=>{:wx=>333, :boundingbox=>[67, 512, 266, 711]}, 203=>{:wx=>333, :boundingbox=>[52, -215, 261, 0]}, 205=>{:wx=>333, :boundingbox=>[-3, 507, 377, 678]}, 206=>{:wx=>333, :boundingbox=>[62, -165, 243, 0]}, 207=>{:wx=>333, :boundingbox=>[11, 507, 322, 674]}, 208=>{:wx=>1000, :boundingbox=>[0, 201, 1000, 250]}, 225=>{:wx=>889, :boundingbox=>[0, 0, 863, 662]}, 227=>{:wx=>276, :boundingbox=>[4, 394, 270, 676]}, 232=>{:wx=>611, :boundingbox=>[12, 0, 598, 662]}, 233=>{:wx=>722, :boundingbox=>[34, -80, 688, 734]}, 234=>{:wx=>889, :boundingbox=>[30, -6, 885, 668]}, 235=>{:wx=>310, :boundingbox=>[6, 394, 304, 676]}, 241=>{:wx=>667, :boundingbox=>[38, -10, 632, 460]}, 245=>{:wx=>278, :boundingbox=>[16, 0, 253, 460]}, 248=>{:wx=>278, :boundingbox=>[19, 0, 259, 683]}, 249=>{:wx=>500, :boundingbox=>[29, -112, 470, 551]}, 250=>{:wx=>722, :boundingbox=>[30, -10, 690, 460]}, 251=>{:wx=>500, :boundingbox=>[12, -9, 468, 683]}, -1=>{:wx=>500, :boundingbox=>[0, 0, 0, 0]}}
			times_bold_metrics = {32=>{:wx=>250, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>333, :boundingbox=>[81, -13, 251, 691]}, 34=>{:wx=>555, :boundingbox=>[83, 404, 472, 691]}, 35=>{:wx=>500, :boundingbox=>[4, 0, 496, 700]}, 36=>{:wx=>500, :boundingbox=>[29, -99, 472, 750]}, 37=>{:wx=>1000, :boundingbox=>[124, -14, 877, 692]}, 38=>{:wx=>833, :boundingbox=>[62, -16, 787, 691]}, 39=>{:wx=>333, :boundingbox=>[79, 356, 263, 691]}, 40=>{:wx=>333, :boundingbox=>[46, -168, 306, 694]}, 41=>{:wx=>333, :boundingbox=>[27, -168, 287, 694]}, 42=>{:wx=>500, :boundingbox=>[56, 255, 447, 691]}, 43=>{:wx=>570, :boundingbox=>[33, 0, 537, 506]}, 44=>{:wx=>250, :boundingbox=>[39, -180, 223, 155]}, 45=>{:wx=>333, :boundingbox=>[44, 171, 287, 287]}, 46=>{:wx=>250, :boundingbox=>[41, -13, 210, 156]}, 47=>{:wx=>278, :boundingbox=>[-24, -19, 302, 691]}, 48=>{:wx=>500, :boundingbox=>[24, -13, 476, 688]}, 49=>{:wx=>500, :boundingbox=>[65, 0, 442, 688]}, 50=>{:wx=>500, :boundingbox=>[17, 0, 478, 688]}, 51=>{:wx=>500, :boundingbox=>[16, -14, 468, 688]}, 52=>{:wx=>500, :boundingbox=>[19, 0, 475, 688]}, 53=>{:wx=>500, :boundingbox=>[22, -8, 470, 676]}, 54=>{:wx=>500, :boundingbox=>[28, -13, 475, 688]}, 55=>{:wx=>500, :boundingbox=>[17, 0, 477, 676]}, 56=>{:wx=>500, :boundingbox=>[28, -13, 472, 688]}, 57=>{:wx=>500, :boundingbox=>[26, -13, 473, 688]}, 58=>{:wx=>333, :boundingbox=>[82, -13, 251, 472]}, 59=>{:wx=>333, :boundingbox=>[82, -180, 266, 472]}, 60=>{:wx=>570, :boundingbox=>[31, -8, 539, 514]}, 61=>{:wx=>570, :boundingbox=>[33, 107, 537, 399]}, 62=>{:wx=>570, :boundingbox=>[31, -8, 539, 514]}, 63=>{:wx=>500, :boundingbox=>[57, -13, 445, 689]}, 64=>{:wx=>930, :boundingbox=>[108, -19, 822, 691]}, 65=>{:wx=>722, :boundingbox=>[9, 0, 689, 690]}, 66=>{:wx=>667, :boundingbox=>[16, 0, 619, 676]}, 67=>{:wx=>722, :boundingbox=>[49, -19, 687, 691]}, 68=>{:wx=>722, :boundingbox=>[14, 0, 690, 676]}, 69=>{:wx=>667, :boundingbox=>[16, 0, 641, 676]}, 70=>{:wx=>611, :boundingbox=>[16, 0, 583, 676]}, 71=>{:wx=>778, :boundingbox=>[37, -19, 755, 691]}, 72=>{:wx=>778, :boundingbox=>[21, 0, 759, 676]}, 73=>{:wx=>389, :boundingbox=>[20, 0, 370, 676]}, 74=>{:wx=>500, :boundingbox=>[3, -96, 479, 676]}, 75=>{:wx=>778, :boundingbox=>[30, 0, 769, 676]}, 76=>{:wx=>667, :boundingbox=>[19, 0, 638, 676]}, 77=>{:wx=>944, :boundingbox=>[14, 0, 921, 676]}, 78=>{:wx=>722, :boundingbox=>[16, -18, 701, 676]}, 79=>{:wx=>778, :boundingbox=>[35, -19, 743, 691]}, 80=>{:wx=>611, :boundingbox=>[16, 0, 600, 676]}, 81=>{:wx=>778, :boundingbox=>[35, -176, 743, 691]}, 82=>{:wx=>722, :boundingbox=>[26, 0, 715, 676]}, 83=>{:wx=>556, :boundingbox=>[35, -19, 513, 692]}, 84=>{:wx=>667, :boundingbox=>[31, 0, 636, 676]}, 85=>{:wx=>722, :boundingbox=>[16, -19, 701, 676]}, 86=>{:wx=>722, :boundingbox=>[16, -18, 701, 676]}, 87=>{:wx=>1000, :boundingbox=>[19, -15, 981, 676]}, 88=>{:wx=>722, :boundingbox=>[16, 0, 699, 676]}, 89=>{:wx=>722, :boundingbox=>[15, 0, 699, 676]}, 90=>{:wx=>667, :boundingbox=>[28, 0, 634, 676]}, 91=>{:wx=>333, :boundingbox=>[67, -149, 301, 678]}, 92=>{:wx=>278, :boundingbox=>[-25, -19, 303, 691]}, 93=>{:wx=>333, :boundingbox=>[32, -149, 266, 678]}, 94=>{:wx=>581, :boundingbox=>[73, 311, 509, 676]}, 95=>{:wx=>500, :boundingbox=>[0, -125, 500, -75]}, 96=>{:wx=>333, :boundingbox=>[70, 356, 254, 691]}, 97=>{:wx=>500, :boundingbox=>[25, -14, 488, 473]}, 98=>{:wx=>556, :boundingbox=>[17, -14, 521, 676]}, 99=>{:wx=>444, :boundingbox=>[25, -14, 430, 473]}, 100=>{:wx=>556, :boundingbox=>[25, -14, 534, 676]}, 101=>{:wx=>444, :boundingbox=>[25, -14, 426, 473]}, 102=>{:wx=>333, :boundingbox=>[14, 0, 389, 691]}, 103=>{:wx=>500, :boundingbox=>[28, -206, 483, 473]}, 104=>{:wx=>556, :boundingbox=>[16, 0, 534, 676]}, 105=>{:wx=>278, :boundingbox=>[16, 0, 255, 691]}, 106=>{:wx=>333, :boundingbox=>[-57, -203, 263, 691]}, 107=>{:wx=>556, :boundingbox=>[22, 0, 543, 676]}, 108=>{:wx=>278, :boundingbox=>[16, 0, 255, 676]}, 109=>{:wx=>833, :boundingbox=>[16, 0, 814, 473]}, 110=>{:wx=>556, :boundingbox=>[21, 0, 539, 473]}, 111=>{:wx=>500, :boundingbox=>[25, -14, 476, 473]}, 112=>{:wx=>556, :boundingbox=>[19, -205, 524, 473]}, 113=>{:wx=>556, :boundingbox=>[34, -205, 536, 473]}, 114=>{:wx=>444, :boundingbox=>[29, 0, 434, 473]}, 115=>{:wx=>389, :boundingbox=>[25, -14, 361, 473]}, 116=>{:wx=>333, :boundingbox=>[20, -12, 332, 630]}, 117=>{:wx=>556, :boundingbox=>[16, -14, 537, 461]}, 118=>{:wx=>500, :boundingbox=>[21, -14, 485, 461]}, 119=>{:wx=>722, :boundingbox=>[23, -14, 707, 461]}, 120=>{:wx=>500, :boundingbox=>[12, 0, 484, 461]}, 121=>{:wx=>500, :boundingbox=>[16, -205, 480, 461]}, 122=>{:wx=>444, :boundingbox=>[21, 0, 420, 461]}, 123=>{:wx=>394, :boundingbox=>[22, -175, 340, 698]}, 124=>{:wx=>220, :boundingbox=>[66, -218, 154, 782]}, 125=>{:wx=>394, :boundingbox=>[54, -175, 372, 698]}, 126=>{:wx=>520, :boundingbox=>[29, 173, 491, 333]}, 161=>{:wx=>333, :boundingbox=>[82, -203, 252, 501]}, 162=>{:wx=>500, :boundingbox=>[53, -140, 458, 588]}, 163=>{:wx=>500, :boundingbox=>[21, -14, 477, 684]}, 164=>{:wx=>167, :boundingbox=>[-168, -12, 329, 688]}, 165=>{:wx=>500, :boundingbox=>[-64, 0, 547, 676]}, 166=>{:wx=>500, :boundingbox=>[0, -155, 498, 706]}, 167=>{:wx=>500, :boundingbox=>[57, -132, 443, 691]}, 168=>{:wx=>500, :boundingbox=>[-26, 61, 526, 613]}, 169=>{:wx=>278, :boundingbox=>[75, 404, 204, 691]}, 170=>{:wx=>500, :boundingbox=>[32, 356, 486, 691]}, 171=>{:wx=>500, :boundingbox=>[23, 36, 473, 415]}, 172=>{:wx=>333, :boundingbox=>[51, 36, 305, 415]}, 173=>{:wx=>333, :boundingbox=>[28, 36, 282, 415]}, 174=>{:wx=>556, :boundingbox=>[14, 0, 536, 691]}, 175=>{:wx=>556, :boundingbox=>[14, 0, 536, 691]}, 177=>{:wx=>500, :boundingbox=>[0, 181, 500, 271]}, 178=>{:wx=>500, :boundingbox=>[47, -134, 453, 691]}, 179=>{:wx=>500, :boundingbox=>[45, -132, 456, 691]}, 180=>{:wx=>250, :boundingbox=>[41, 248, 210, 417]}, 182=>{:wx=>540, :boundingbox=>[0, -186, 519, 676]}, 183=>{:wx=>350, :boundingbox=>[35, 198, 315, 478]}, 184=>{:wx=>333, :boundingbox=>[79, -180, 263, 155]}, 185=>{:wx=>500, :boundingbox=>[14, -180, 468, 155]}, 186=>{:wx=>500, :boundingbox=>[14, 356, 468, 691]}, 187=>{:wx=>500, :boundingbox=>[27, 36, 477, 415]}, 188=>{:wx=>1000, :boundingbox=>[82, -13, 917, 156]}, 189=>{:wx=>1000, :boundingbox=>[7, -29, 995, 706]}, 191=>{:wx=>500, :boundingbox=>[55, -201, 443, 501]}, 193=>{:wx=>333, :boundingbox=>[8, 528, 246, 713]}, 194=>{:wx=>333, :boundingbox=>[86, 528, 324, 713]}, 195=>{:wx=>333, :boundingbox=>[-2, 528, 335, 704]}, 196=>{:wx=>333, :boundingbox=>[-16, 547, 349, 674]}, 197=>{:wx=>333, :boundingbox=>[1, 565, 331, 637]}, 198=>{:wx=>333, :boundingbox=>[15, 528, 318, 691]}, 199=>{:wx=>333, :boundingbox=>[103, 536, 258, 691]}, 200=>{:wx=>333, :boundingbox=>[-2, 537, 335, 667]}, 202=>{:wx=>333, :boundingbox=>[60, 527, 273, 740]}, 203=>{:wx=>333, :boundingbox=>[68, -218, 294, 0]}, 205=>{:wx=>333, :boundingbox=>[-13, 528, 425, 713]}, 206=>{:wx=>333, :boundingbox=>[90, -193, 319, 24]}, 207=>{:wx=>333, :boundingbox=>[-2, 528, 335, 704]}, 208=>{:wx=>1000, :boundingbox=>[0, 181, 1000, 271]}, 225=>{:wx=>1000, :boundingbox=>[4, 0, 951, 676]}, 227=>{:wx=>300, :boundingbox=>[-1, 397, 301, 688]}, 232=>{:wx=>667, :boundingbox=>[19, 0, 638, 676]}, 233=>{:wx=>778, :boundingbox=>[35, -74, 743, 737]}, 234=>{:wx=>1000, :boundingbox=>[22, -5, 981, 684]}, 235=>{:wx=>330, :boundingbox=>[18, 397, 312, 688]}, 241=>{:wx=>722, :boundingbox=>[33, -14, 693, 473]}, 245=>{:wx=>278, :boundingbox=>[16, 0, 255, 461]}, 248=>{:wx=>278, :boundingbox=>[-22, 0, 303, 676]}, 249=>{:wx=>500, :boundingbox=>[25, -92, 476, 549]}, 250=>{:wx=>722, :boundingbox=>[22, -14, 696, 473]}, 251=>{:wx=>556, :boundingbox=>[19, -12, 517, 691]}, -1=>{:wx=>500, :boundingbox=>[0, 0, 0, 0]}}
			times_italic_metrics = {32=>{:wx=>250, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>333, :boundingbox=>[39, -11, 302, 667]}, 34=>{:wx=>420, :boundingbox=>[144, 421, 432, 666]}, 35=>{:wx=>500, :boundingbox=>[2, 0, 540, 676]}, 36=>{:wx=>500, :boundingbox=>[31, -89, 497, 731]}, 37=>{:wx=>833, :boundingbox=>[79, -13, 790, 676]}, 38=>{:wx=>778, :boundingbox=>[76, -18, 723, 666]}, 39=>{:wx=>333, :boundingbox=>[151, 436, 290, 666]}, 40=>{:wx=>333, :boundingbox=>[42, -181, 315, 669]}, 41=>{:wx=>333, :boundingbox=>[16, -180, 289, 669]}, 42=>{:wx=>500, :boundingbox=>[128, 255, 492, 666]}, 43=>{:wx=>675, :boundingbox=>[86, 0, 590, 506]}, 44=>{:wx=>250, :boundingbox=>[-4, -129, 135, 101]}, 45=>{:wx=>333, :boundingbox=>[49, 192, 282, 255]}, 46=>{:wx=>250, :boundingbox=>[27, -11, 138, 100]}, 47=>{:wx=>278, :boundingbox=>[-65, -18, 386, 666]}, 48=>{:wx=>500, :boundingbox=>[32, -7, 497, 676]}, 49=>{:wx=>500, :boundingbox=>[49, 0, 409, 676]}, 50=>{:wx=>500, :boundingbox=>[12, 0, 452, 676]}, 51=>{:wx=>500, :boundingbox=>[15, -7, 465, 676]}, 52=>{:wx=>500, :boundingbox=>[1, 0, 479, 676]}, 53=>{:wx=>500, :boundingbox=>[15, -7, 491, 666]}, 54=>{:wx=>500, :boundingbox=>[30, -7, 521, 686]}, 55=>{:wx=>500, :boundingbox=>[75, -8, 537, 666]}, 56=>{:wx=>500, :boundingbox=>[30, -7, 493, 676]}, 57=>{:wx=>500, :boundingbox=>[23, -17, 492, 676]}, 58=>{:wx=>333, :boundingbox=>[50, -11, 261, 441]}, 59=>{:wx=>333, :boundingbox=>[27, -129, 261, 441]}, 60=>{:wx=>675, :boundingbox=>[84, -8, 592, 514]}, 61=>{:wx=>675, :boundingbox=>[86, 120, 590, 386]}, 62=>{:wx=>675, :boundingbox=>[84, -8, 592, 514]}, 63=>{:wx=>500, :boundingbox=>[132, -12, 472, 664]}, 64=>{:wx=>920, :boundingbox=>[118, -18, 806, 666]}, 65=>{:wx=>611, :boundingbox=>[-51, 0, 564, 668]}, 66=>{:wx=>611, :boundingbox=>[-8, 0, 588, 653]}, 67=>{:wx=>667, :boundingbox=>[66, -18, 689, 666]}, 68=>{:wx=>722, :boundingbox=>[-8, 0, 700, 653]}, 69=>{:wx=>611, :boundingbox=>[-1, 0, 634, 653]}, 70=>{:wx=>611, :boundingbox=>[8, 0, 645, 653]}, 71=>{:wx=>722, :boundingbox=>[52, -18, 722, 666]}, 72=>{:wx=>722, :boundingbox=>[-8, 0, 767, 653]}, 73=>{:wx=>333, :boundingbox=>[-8, 0, 384, 653]}, 74=>{:wx=>444, :boundingbox=>[-6, -18, 491, 653]}, 75=>{:wx=>667, :boundingbox=>[7, 0, 722, 653]}, 76=>{:wx=>556, :boundingbox=>[-8, 0, 559, 653]}, 77=>{:wx=>833, :boundingbox=>[-18, 0, 873, 653]}, 78=>{:wx=>667, :boundingbox=>[-20, -15, 727, 653]}, 79=>{:wx=>722, :boundingbox=>[60, -18, 699, 666]}, 80=>{:wx=>611, :boundingbox=>[0, 0, 605, 653]}, 81=>{:wx=>722, :boundingbox=>[59, -182, 699, 666]}, 82=>{:wx=>611, :boundingbox=>[-13, 0, 588, 653]}, 83=>{:wx=>500, :boundingbox=>[17, -18, 508, 667]}, 84=>{:wx=>556, :boundingbox=>[59, 0, 633, 653]}, 85=>{:wx=>722, :boundingbox=>[102, -18, 765, 653]}, 86=>{:wx=>611, :boundingbox=>[76, -18, 688, 653]}, 87=>{:wx=>833, :boundingbox=>[71, -18, 906, 653]}, 88=>{:wx=>611, :boundingbox=>[-29, 0, 655, 653]}, 89=>{:wx=>556, :boundingbox=>[78, 0, 633, 653]}, 90=>{:wx=>556, :boundingbox=>[-6, 0, 606, 653]}, 91=>{:wx=>389, :boundingbox=>[21, -153, 391, 663]}, 92=>{:wx=>278, :boundingbox=>[-41, -18, 319, 666]}, 93=>{:wx=>389, :boundingbox=>[12, -153, 382, 663]}, 94=>{:wx=>422, :boundingbox=>[0, 301, 422, 666]}, 95=>{:wx=>500, :boundingbox=>[0, -125, 500, -75]}, 96=>{:wx=>333, :boundingbox=>[171, 436, 310, 666]}, 97=>{:wx=>500, :boundingbox=>[17, -11, 476, 441]}, 98=>{:wx=>500, :boundingbox=>[23, -11, 473, 683]}, 99=>{:wx=>444, :boundingbox=>[30, -11, 425, 441]}, 100=>{:wx=>500, :boundingbox=>[15, -13, 527, 683]}, 101=>{:wx=>444, :boundingbox=>[31, -11, 412, 441]}, 102=>{:wx=>278, :boundingbox=>[-147, -207, 424, 678]}, 103=>{:wx=>500, :boundingbox=>[8, -206, 472, 441]}, 104=>{:wx=>500, :boundingbox=>[19, -9, 478, 683]}, 105=>{:wx=>278, :boundingbox=>[49, -11, 264, 654]}, 106=>{:wx=>278, :boundingbox=>[-124, -207, 276, 654]}, 107=>{:wx=>444, :boundingbox=>[14, -11, 461, 683]}, 108=>{:wx=>278, :boundingbox=>[41, -11, 279, 683]}, 109=>{:wx=>722, :boundingbox=>[12, -9, 704, 441]}, 110=>{:wx=>500, :boundingbox=>[14, -9, 474, 441]}, 111=>{:wx=>500, :boundingbox=>[27, -11, 468, 441]}, 112=>{:wx=>500, :boundingbox=>[-75, -205, 469, 441]}, 113=>{:wx=>500, :boundingbox=>[25, -209, 483, 441]}, 114=>{:wx=>389, :boundingbox=>[45, 0, 412, 441]}, 115=>{:wx=>389, :boundingbox=>[16, -13, 366, 442]}, 116=>{:wx=>278, :boundingbox=>[37, -11, 296, 546]}, 117=>{:wx=>500, :boundingbox=>[42, -11, 475, 441]}, 118=>{:wx=>444, :boundingbox=>[21, -18, 426, 441]}, 119=>{:wx=>667, :boundingbox=>[16, -18, 648, 441]}, 120=>{:wx=>444, :boundingbox=>[-27, -11, 447, 441]}, 121=>{:wx=>444, :boundingbox=>[-24, -206, 426, 441]}, 122=>{:wx=>389, :boundingbox=>[-2, -81, 380, 428]}, 123=>{:wx=>400, :boundingbox=>[51, -177, 407, 687]}, 124=>{:wx=>275, :boundingbox=>[105, -217, 171, 783]}, 125=>{:wx=>400, :boundingbox=>[-7, -177, 349, 687]}, 126=>{:wx=>541, :boundingbox=>[40, 183, 502, 323]}, 161=>{:wx=>389, :boundingbox=>[59, -205, 322, 473]}, 162=>{:wx=>500, :boundingbox=>[77, -143, 472, 560]}, 163=>{:wx=>500, :boundingbox=>[10, -6, 517, 670]}, 164=>{:wx=>167, :boundingbox=>[-169, -10, 337, 676]}, 165=>{:wx=>500, :boundingbox=>[27, 0, 603, 653]}, 166=>{:wx=>500, :boundingbox=>[25, -182, 507, 682]}, 167=>{:wx=>500, :boundingbox=>[53, -162, 461, 666]}, 168=>{:wx=>500, :boundingbox=>[-22, 53, 522, 597]}, 169=>{:wx=>214, :boundingbox=>[132, 421, 241, 666]}, 170=>{:wx=>556, :boundingbox=>[166, 436, 514, 666]}, 171=>{:wx=>500, :boundingbox=>[53, 37, 445, 403]}, 172=>{:wx=>333, :boundingbox=>[51, 37, 281, 403]}, 173=>{:wx=>333, :boundingbox=>[52, 37, 282, 403]}, 174=>{:wx=>500, :boundingbox=>[-141, -207, 481, 681]}, 175=>{:wx=>500, :boundingbox=>[-141, -204, 518, 682]}, 177=>{:wx=>500, :boundingbox=>[-6, 197, 505, 243]}, 178=>{:wx=>500, :boundingbox=>[101, -159, 488, 666]}, 179=>{:wx=>500, :boundingbox=>[22, -143, 491, 666]}, 180=>{:wx=>250, :boundingbox=>[70, 199, 181, 310]}, 182=>{:wx=>523, :boundingbox=>[55, -123, 616, 653]}, 183=>{:wx=>350, :boundingbox=>[40, 191, 310, 461]}, 184=>{:wx=>333, :boundingbox=>[44, -129, 183, 101]}, 185=>{:wx=>556, :boundingbox=>[57, -129, 405, 101]}, 186=>{:wx=>556, :boundingbox=>[151, 436, 499, 666]}, 187=>{:wx=>500, :boundingbox=>[55, 37, 447, 403]}, 188=>{:wx=>889, :boundingbox=>[57, -11, 762, 100]}, 189=>{:wx=>1000, :boundingbox=>[25, -19, 1010, 706]}, 191=>{:wx=>500, :boundingbox=>[28, -205, 368, 471]}, 193=>{:wx=>333, :boundingbox=>[121, 492, 311, 664]}, 194=>{:wx=>333, :boundingbox=>[180, 494, 403, 664]}, 195=>{:wx=>333, :boundingbox=>[91, 492, 385, 661]}, 196=>{:wx=>333, :boundingbox=>[100, 517, 427, 624]}, 197=>{:wx=>333, :boundingbox=>[99, 532, 411, 583]}, 198=>{:wx=>333, :boundingbox=>[117, 492, 418, 650]}, 199=>{:wx=>333, :boundingbox=>[207, 548, 305, 646]}, 200=>{:wx=>333, :boundingbox=>[107, 548, 405, 646]}, 202=>{:wx=>333, :boundingbox=>[155, 492, 355, 691]}, 203=>{:wx=>333, :boundingbox=>[-30, -217, 182, 0]}, 205=>{:wx=>333, :boundingbox=>[93, 494, 486, 664]}, 206=>{:wx=>333, :boundingbox=>[20, -169, 203, 40]}, 207=>{:wx=>333, :boundingbox=>[121, 492, 426, 661]}, 208=>{:wx=>889, :boundingbox=>[-6, 197, 894, 243]}, 225=>{:wx=>889, :boundingbox=>[-27, 0, 911, 653]}, 227=>{:wx=>276, :boundingbox=>[42, 406, 352, 676]}, 232=>{:wx=>556, :boundingbox=>[-8, 0, 559, 653]}, 233=>{:wx=>722, :boundingbox=>[60, -105, 699, 722]}, 234=>{:wx=>944, :boundingbox=>[49, -8, 964, 666]}, 235=>{:wx=>310, :boundingbox=>[67, 406, 362, 676]}, 241=>{:wx=>667, :boundingbox=>[23, -11, 640, 441]}, 245=>{:wx=>278, :boundingbox=>[49, -11, 235, 441]}, 248=>{:wx=>278, :boundingbox=>[41, -11, 312, 683]}, 249=>{:wx=>500, :boundingbox=>[28, -135, 469, 554]}, 250=>{:wx=>667, :boundingbox=>[20, -12, 646, 441]}, 251=>{:wx=>500, :boundingbox=>[-168, -207, 493, 679]}, -1=>{:wx=>500, :boundingbox=>[0, 0, 0, 0]}}
			times_bolditalic_metrics = {32=>{:wx=>250, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>389, :boundingbox=>[67, -13, 370, 684]}, 34=>{:wx=>555, :boundingbox=>[136, 398, 536, 685]}, 35=>{:wx=>500, :boundingbox=>[-33, 0, 533, 700]}, 36=>{:wx=>500, :boundingbox=>[-20, -100, 497, 733]}, 37=>{:wx=>833, :boundingbox=>[39, -10, 793, 692]}, 38=>{:wx=>778, :boundingbox=>[5, -19, 699, 682]}, 39=>{:wx=>333, :boundingbox=>[98, 369, 302, 685]}, 40=>{:wx=>333, :boundingbox=>[28, -179, 344, 685]}, 41=>{:wx=>333, :boundingbox=>[-44, -179, 271, 685]}, 42=>{:wx=>500, :boundingbox=>[65, 249, 456, 685]}, 43=>{:wx=>570, :boundingbox=>[33, 0, 537, 506]}, 44=>{:wx=>250, :boundingbox=>[-60, -182, 144, 134]}, 45=>{:wx=>333, :boundingbox=>[2, 166, 271, 282]}, 46=>{:wx=>250, :boundingbox=>[-9, -13, 139, 135]}, 47=>{:wx=>278, :boundingbox=>[-64, -18, 342, 685]}, 48=>{:wx=>500, :boundingbox=>[17, -14, 477, 683]}, 49=>{:wx=>500, :boundingbox=>[5, 0, 419, 683]}, 50=>{:wx=>500, :boundingbox=>[-27, 0, 446, 683]}, 51=>{:wx=>500, :boundingbox=>[-15, -13, 450, 683]}, 52=>{:wx=>500, :boundingbox=>[-15, 0, 503, 683]}, 53=>{:wx=>500, :boundingbox=>[-11, -13, 487, 669]}, 54=>{:wx=>500, :boundingbox=>[23, -15, 509, 679]}, 55=>{:wx=>500, :boundingbox=>[52, 0, 525, 669]}, 56=>{:wx=>500, :boundingbox=>[3, -13, 476, 683]}, 57=>{:wx=>500, :boundingbox=>[-12, -10, 475, 683]}, 58=>{:wx=>333, :boundingbox=>[23, -13, 264, 459]}, 59=>{:wx=>333, :boundingbox=>[-25, -183, 264, 459]}, 60=>{:wx=>570, :boundingbox=>[31, -8, 539, 514]}, 61=>{:wx=>570, :boundingbox=>[33, 107, 537, 399]}, 62=>{:wx=>570, :boundingbox=>[31, -8, 539, 514]}, 63=>{:wx=>500, :boundingbox=>[79, -13, 470, 684]}, 64=>{:wx=>832, :boundingbox=>[63, -18, 770, 685]}, 65=>{:wx=>667, :boundingbox=>[-67, 0, 593, 683]}, 66=>{:wx=>667, :boundingbox=>[-24, 0, 624, 669]}, 67=>{:wx=>667, :boundingbox=>[32, -18, 677, 685]}, 68=>{:wx=>722, :boundingbox=>[-46, 0, 685, 669]}, 69=>{:wx=>667, :boundingbox=>[-27, 0, 653, 669]}, 70=>{:wx=>667, :boundingbox=>[-13, 0, 660, 669]}, 71=>{:wx=>722, :boundingbox=>[21, -18, 706, 685]}, 72=>{:wx=>778, :boundingbox=>[-24, 0, 799, 669]}, 73=>{:wx=>389, :boundingbox=>[-32, 0, 406, 669]}, 74=>{:wx=>500, :boundingbox=>[-46, -99, 524, 669]}, 75=>{:wx=>667, :boundingbox=>[-21, 0, 702, 669]}, 76=>{:wx=>611, :boundingbox=>[-22, 0, 590, 669]}, 77=>{:wx=>889, :boundingbox=>[-29, -12, 917, 669]}, 78=>{:wx=>722, :boundingbox=>[-27, -15, 748, 669]}, 79=>{:wx=>722, :boundingbox=>[27, -18, 691, 685]}, 80=>{:wx=>611, :boundingbox=>[-27, 0, 613, 669]}, 81=>{:wx=>722, :boundingbox=>[27, -208, 691, 685]}, 82=>{:wx=>667, :boundingbox=>[-29, 0, 623, 669]}, 83=>{:wx=>556, :boundingbox=>[2, -18, 526, 685]}, 84=>{:wx=>611, :boundingbox=>[50, 0, 650, 669]}, 85=>{:wx=>722, :boundingbox=>[67, -18, 744, 669]}, 86=>{:wx=>667, :boundingbox=>[65, -18, 715, 669]}, 87=>{:wx=>889, :boundingbox=>[65, -18, 940, 669]}, 88=>{:wx=>667, :boundingbox=>[-24, 0, 694, 669]}, 89=>{:wx=>611, :boundingbox=>[73, 0, 659, 669]}, 90=>{:wx=>611, :boundingbox=>[-11, 0, 590, 669]}, 91=>{:wx=>333, :boundingbox=>[-37, -159, 362, 674]}, 92=>{:wx=>278, :boundingbox=>[-1, -18, 279, 685]}, 93=>{:wx=>333, :boundingbox=>[-56, -157, 343, 674]}, 94=>{:wx=>570, :boundingbox=>[67, 304, 503, 669]}, 95=>{:wx=>500, :boundingbox=>[0, -125, 500, -75]}, 96=>{:wx=>333, :boundingbox=>[128, 369, 332, 685]}, 97=>{:wx=>500, :boundingbox=>[-21, -14, 455, 462]}, 98=>{:wx=>500, :boundingbox=>[-14, -13, 444, 699]}, 99=>{:wx=>444, :boundingbox=>[-5, -13, 392, 462]}, 100=>{:wx=>500, :boundingbox=>[-21, -13, 517, 699]}, 101=>{:wx=>444, :boundingbox=>[5, -13, 398, 462]}, 102=>{:wx=>333, :boundingbox=>[-169, -205, 446, 698]}, 103=>{:wx=>500, :boundingbox=>[-52, -203, 478, 462]}, 104=>{:wx=>556, :boundingbox=>[-13, -9, 498, 699]}, 105=>{:wx=>278, :boundingbox=>[2, -9, 263, 684]}, 106=>{:wx=>278, :boundingbox=>[-189, -207, 279, 684]}, 107=>{:wx=>500, :boundingbox=>[-23, -8, 483, 699]}, 108=>{:wx=>278, :boundingbox=>[2, -9, 290, 699]}, 109=>{:wx=>778, :boundingbox=>[-14, -9, 722, 462]}, 110=>{:wx=>556, :boundingbox=>[-6, -9, 493, 462]}, 111=>{:wx=>500, :boundingbox=>[-3, -13, 441, 462]}, 112=>{:wx=>500, :boundingbox=>[-120, -205, 446, 462]}, 113=>{:wx=>500, :boundingbox=>[1, -205, 471, 462]}, 114=>{:wx=>389, :boundingbox=>[-21, 0, 389, 462]}, 115=>{:wx=>389, :boundingbox=>[-19, -13, 333, 462]}, 116=>{:wx=>278, :boundingbox=>[-11, -9, 281, 594]}, 117=>{:wx=>556, :boundingbox=>[15, -9, 492, 462]}, 118=>{:wx=>444, :boundingbox=>[16, -13, 401, 462]}, 119=>{:wx=>667, :boundingbox=>[16, -13, 614, 462]}, 120=>{:wx=>500, :boundingbox=>[-46, -13, 469, 462]}, 121=>{:wx=>444, :boundingbox=>[-94, -205, 392, 462]}, 122=>{:wx=>389, :boundingbox=>[-43, -78, 368, 449]}, 123=>{:wx=>348, :boundingbox=>[5, -187, 436, 686]}, 124=>{:wx=>220, :boundingbox=>[66, -218, 154, 782]}, 125=>{:wx=>348, :boundingbox=>[-129, -187, 302, 686]}, 126=>{:wx=>570, :boundingbox=>[54, 173, 516, 333]}, 161=>{:wx=>389, :boundingbox=>[19, -205, 322, 492]}, 162=>{:wx=>500, :boundingbox=>[42, -143, 439, 576]}, 163=>{:wx=>500, :boundingbox=>[-32, -12, 510, 683]}, 164=>{:wx=>167, :boundingbox=>[-169, -14, 324, 683]}, 165=>{:wx=>500, :boundingbox=>[33, 0, 628, 669]}, 166=>{:wx=>500, :boundingbox=>[-87, -156, 537, 707]}, 167=>{:wx=>500, :boundingbox=>[36, -143, 459, 685]}, 168=>{:wx=>500, :boundingbox=>[-26, 34, 526, 586]}, 169=>{:wx=>278, :boundingbox=>[128, 398, 268, 685]}, 170=>{:wx=>500, :boundingbox=>[53, 369, 513, 685]}, 171=>{:wx=>500, :boundingbox=>[12, 32, 468, 415]}, 172=>{:wx=>333, :boundingbox=>[32, 32, 303, 415]}, 173=>{:wx=>333, :boundingbox=>[10, 32, 281, 415]}, 174=>{:wx=>556, :boundingbox=>[-188, -205, 514, 703]}, 175=>{:wx=>556, :boundingbox=>[-186, -205, 553, 704]}, 177=>{:wx=>500, :boundingbox=>[-40, 178, 477, 269]}, 178=>{:wx=>500, :boundingbox=>[91, -145, 494, 685]}, 179=>{:wx=>500, :boundingbox=>[10, -139, 493, 685]}, 180=>{:wx=>250, :boundingbox=>[51, 257, 199, 405]}, 182=>{:wx=>500, :boundingbox=>[-57, -193, 562, 669]}, 183=>{:wx=>350, :boundingbox=>[0, 175, 350, 525]}, 184=>{:wx=>333, :boundingbox=>[-5, -182, 199, 134]}, 185=>{:wx=>500, :boundingbox=>[-57, -182, 403, 134]}, 186=>{:wx=>500, :boundingbox=>[53, 369, 513, 685]}, 187=>{:wx=>500, :boundingbox=>[12, 32, 468, 415]}, 188=>{:wx=>1000, :boundingbox=>[40, -13, 852, 135]}, 189=>{:wx=>1000, :boundingbox=>[7, -29, 996, 706]}, 191=>{:wx=>500, :boundingbox=>[30, -205, 421, 492]}, 193=>{:wx=>333, :boundingbox=>[85, 516, 297, 697]}, 194=>{:wx=>333, :boundingbox=>[139, 516, 379, 697]}, 195=>{:wx=>333, :boundingbox=>[40, 516, 367, 690]}, 196=>{:wx=>333, :boundingbox=>[48, 536, 407, 655]}, 197=>{:wx=>333, :boundingbox=>[51, 553, 393, 623]}, 198=>{:wx=>333, :boundingbox=>[71, 516, 387, 678]}, 199=>{:wx=>333, :boundingbox=>[163, 550, 298, 684]}, 200=>{:wx=>333, :boundingbox=>[55, 550, 402, 684]}, 202=>{:wx=>333, :boundingbox=>[127, 516, 340, 729]}, 203=>{:wx=>333, :boundingbox=>[-80, -218, 156, 5]}, 205=>{:wx=>333, :boundingbox=>[69, 516, 498, 697]}, 206=>{:wx=>333, :boundingbox=>[15, -183, 244, 34]}, 207=>{:wx=>333, :boundingbox=>[79, 516, 411, 690]}, 208=>{:wx=>1000, :boundingbox=>[-40, 178, 977, 269]}, 225=>{:wx=>944, :boundingbox=>[-64, 0, 918, 669]}, 227=>{:wx=>266, :boundingbox=>[16, 399, 330, 685]}, 232=>{:wx=>611, :boundingbox=>[-22, 0, 590, 669]}, 233=>{:wx=>722, :boundingbox=>[27, -125, 691, 764]}, 234=>{:wx=>944, :boundingbox=>[23, -8, 946, 677]}, 235=>{:wx=>300, :boundingbox=>[56, 400, 347, 685]}, 241=>{:wx=>722, :boundingbox=>[-5, -13, 673, 462]}, 245=>{:wx=>278, :boundingbox=>[2, -9, 238, 462]}, 248=>{:wx=>278, :boundingbox=>[-7, -9, 307, 699]}, 249=>{:wx=>500, :boundingbox=>[-3, -119, 441, 560]}, 250=>{:wx=>722, :boundingbox=>[6, -13, 674, 462]}, 251=>{:wx=>500, :boundingbox=>[-200, -200, 473, 705]}, -1=>{:wx=>500, :boundingbox=>[0, 0, 0, 0]}}
			helvetica_metrics = {32=>{:wx=>278, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>278, :boundingbox=>[90, 0, 187, 718]}, 34=>{:wx=>355, :boundingbox=>[70, 463, 285, 718]}, 35=>{:wx=>556, :boundingbox=>[28, 0, 529, 688]}, 36=>{:wx=>556, :boundingbox=>[32, -115, 520, 775]}, 37=>{:wx=>889, :boundingbox=>[39, -19, 850, 703]}, 38=>{:wx=>667, :boundingbox=>[44, -15, 645, 718]}, 39=>{:wx=>222, :boundingbox=>[53, 463, 157, 718]}, 40=>{:wx=>333, :boundingbox=>[68, -207, 299, 733]}, 41=>{:wx=>333, :boundingbox=>[34, -207, 265, 733]}, 42=>{:wx=>389, :boundingbox=>[39, 431, 349, 718]}, 43=>{:wx=>584, :boundingbox=>[39, 0, 545, 505]}, 44=>{:wx=>278, :boundingbox=>[87, -147, 191, 106]}, 45=>{:wx=>333, :boundingbox=>[44, 232, 289, 322]}, 46=>{:wx=>278, :boundingbox=>[87, 0, 191, 106]}, 47=>{:wx=>278, :boundingbox=>[-17, -19, 295, 737]}, 48=>{:wx=>556, :boundingbox=>[37, -19, 519, 703]}, 49=>{:wx=>556, :boundingbox=>[101, 0, 359, 703]}, 50=>{:wx=>556, :boundingbox=>[26, 0, 507, 703]}, 51=>{:wx=>556, :boundingbox=>[34, -19, 522, 703]}, 52=>{:wx=>556, :boundingbox=>[25, 0, 523, 703]}, 53=>{:wx=>556, :boundingbox=>[32, -19, 514, 688]}, 54=>{:wx=>556, :boundingbox=>[38, -19, 518, 703]}, 55=>{:wx=>556, :boundingbox=>[37, 0, 523, 688]}, 56=>{:wx=>556, :boundingbox=>[38, -19, 517, 703]}, 57=>{:wx=>556, :boundingbox=>[42, -19, 514, 703]}, 58=>{:wx=>278, :boundingbox=>[87, 0, 191, 516]}, 59=>{:wx=>278, :boundingbox=>[87, -147, 191, 516]}, 60=>{:wx=>584, :boundingbox=>[48, 11, 536, 495]}, 61=>{:wx=>584, :boundingbox=>[39, 115, 545, 390]}, 62=>{:wx=>584, :boundingbox=>[48, 11, 536, 495]}, 63=>{:wx=>556, :boundingbox=>[56, 0, 492, 727]}, 64=>{:wx=>1015, :boundingbox=>[147, -19, 868, 737]}, 65=>{:wx=>667, :boundingbox=>[14, 0, 654, 718]}, 66=>{:wx=>667, :boundingbox=>[74, 0, 627, 718]}, 67=>{:wx=>722, :boundingbox=>[44, -19, 681, 737]}, 68=>{:wx=>722, :boundingbox=>[81, 0, 674, 718]}, 69=>{:wx=>667, :boundingbox=>[86, 0, 616, 718]}, 70=>{:wx=>611, :boundingbox=>[86, 0, 583, 718]}, 71=>{:wx=>778, :boundingbox=>[48, -19, 704, 737]}, 72=>{:wx=>722, :boundingbox=>[77, 0, 646, 718]}, 73=>{:wx=>278, :boundingbox=>[91, 0, 188, 718]}, 74=>{:wx=>500, :boundingbox=>[17, -19, 428, 718]}, 75=>{:wx=>667, :boundingbox=>[76, 0, 663, 718]}, 76=>{:wx=>556, :boundingbox=>[76, 0, 537, 718]}, 77=>{:wx=>833, :boundingbox=>[73, 0, 761, 718]}, 78=>{:wx=>722, :boundingbox=>[76, 0, 646, 718]}, 79=>{:wx=>778, :boundingbox=>[39, -19, 739, 737]}, 80=>{:wx=>667, :boundingbox=>[86, 0, 622, 718]}, 81=>{:wx=>778, :boundingbox=>[39, -56, 739, 737]}, 82=>{:wx=>722, :boundingbox=>[88, 0, 684, 718]}, 83=>{:wx=>667, :boundingbox=>[49, -19, 620, 737]}, 84=>{:wx=>611, :boundingbox=>[14, 0, 597, 718]}, 85=>{:wx=>722, :boundingbox=>[79, -19, 644, 718]}, 86=>{:wx=>667, :boundingbox=>[20, 0, 647, 718]}, 87=>{:wx=>944, :boundingbox=>[16, 0, 928, 718]}, 88=>{:wx=>667, :boundingbox=>[19, 0, 648, 718]}, 89=>{:wx=>667, :boundingbox=>[14, 0, 653, 718]}, 90=>{:wx=>611, :boundingbox=>[23, 0, 588, 718]}, 91=>{:wx=>278, :boundingbox=>[63, -196, 250, 722]}, 92=>{:wx=>278, :boundingbox=>[-17, -19, 295, 737]}, 93=>{:wx=>278, :boundingbox=>[28, -196, 215, 722]}, 94=>{:wx=>469, :boundingbox=>[-14, 264, 483, 688]}, 95=>{:wx=>556, :boundingbox=>[0, -125, 556, -75]}, 96=>{:wx=>222, :boundingbox=>[65, 470, 169, 725]}, 97=>{:wx=>556, :boundingbox=>[36, -15, 530, 538]}, 98=>{:wx=>556, :boundingbox=>[58, -15, 517, 718]}, 99=>{:wx=>500, :boundingbox=>[30, -15, 477, 538]}, 100=>{:wx=>556, :boundingbox=>[35, -15, 499, 718]}, 101=>{:wx=>556, :boundingbox=>[40, -15, 516, 538]}, 102=>{:wx=>278, :boundingbox=>[14, 0, 262, 728]}, 103=>{:wx=>556, :boundingbox=>[40, -220, 499, 538]}, 104=>{:wx=>556, :boundingbox=>[65, 0, 491, 718]}, 105=>{:wx=>222, :boundingbox=>[67, 0, 155, 718]}, 106=>{:wx=>222, :boundingbox=>[-16, -210, 155, 718]}, 107=>{:wx=>500, :boundingbox=>[67, 0, 501, 718]}, 108=>{:wx=>222, :boundingbox=>[67, 0, 155, 718]}, 109=>{:wx=>833, :boundingbox=>[65, 0, 769, 538]}, 110=>{:wx=>556, :boundingbox=>[65, 0, 491, 538]}, 111=>{:wx=>556, :boundingbox=>[35, -14, 521, 538]}, 112=>{:wx=>556, :boundingbox=>[58, -207, 517, 538]}, 113=>{:wx=>556, :boundingbox=>[35, -207, 494, 538]}, 114=>{:wx=>333, :boundingbox=>[77, 0, 332, 538]}, 115=>{:wx=>500, :boundingbox=>[32, -15, 464, 538]}, 116=>{:wx=>278, :boundingbox=>[14, -7, 257, 669]}, 117=>{:wx=>556, :boundingbox=>[68, -15, 489, 523]}, 118=>{:wx=>500, :boundingbox=>[8, 0, 492, 523]}, 119=>{:wx=>722, :boundingbox=>[14, 0, 709, 523]}, 120=>{:wx=>500, :boundingbox=>[11, 0, 490, 523]}, 121=>{:wx=>500, :boundingbox=>[11, -214, 489, 523]}, 122=>{:wx=>500, :boundingbox=>[31, 0, 469, 523]}, 123=>{:wx=>334, :boundingbox=>[42, -196, 292, 722]}, 124=>{:wx=>260, :boundingbox=>[94, -225, 167, 775]}, 125=>{:wx=>334, :boundingbox=>[42, -196, 292, 722]}, 126=>{:wx=>584, :boundingbox=>[61, 180, 523, 326]}, 161=>{:wx=>333, :boundingbox=>[118, -195, 215, 523]}, 162=>{:wx=>556, :boundingbox=>[51, -115, 513, 623]}, 163=>{:wx=>556, :boundingbox=>[33, -16, 539, 718]}, 164=>{:wx=>167, :boundingbox=>[-166, -19, 333, 703]}, 165=>{:wx=>556, :boundingbox=>[3, 0, 553, 688]}, 166=>{:wx=>556, :boundingbox=>[-11, -207, 501, 737]}, 167=>{:wx=>556, :boundingbox=>[43, -191, 512, 737]}, 168=>{:wx=>556, :boundingbox=>[28, 99, 528, 603]}, 169=>{:wx=>191, :boundingbox=>[59, 463, 132, 718]}, 170=>{:wx=>333, :boundingbox=>[38, 470, 307, 725]}, 171=>{:wx=>556, :boundingbox=>[97, 108, 459, 446]}, 172=>{:wx=>333, :boundingbox=>[88, 108, 245, 446]}, 173=>{:wx=>333, :boundingbox=>[88, 108, 245, 446]}, 174=>{:wx=>500, :boundingbox=>[14, 0, 434, 728]}, 175=>{:wx=>500, :boundingbox=>[14, 0, 432, 728]}, 177=>{:wx=>556, :boundingbox=>[0, 240, 556, 313]}, 178=>{:wx=>556, :boundingbox=>[43, -159, 514, 718]}, 179=>{:wx=>556, :boundingbox=>[43, -159, 514, 718]}, 180=>{:wx=>278, :boundingbox=>[77, 190, 202, 315]}, 182=>{:wx=>537, :boundingbox=>[18, -173, 497, 718]}, 183=>{:wx=>350, :boundingbox=>[18, 202, 333, 517]}, 184=>{:wx=>222, :boundingbox=>[53, -149, 157, 106]}, 185=>{:wx=>333, :boundingbox=>[26, -149, 295, 106]}, 186=>{:wx=>333, :boundingbox=>[26, 463, 295, 718]}, 187=>{:wx=>556, :boundingbox=>[97, 108, 459, 446]}, 188=>{:wx=>1000, :boundingbox=>[115, 0, 885, 106]}, 189=>{:wx=>1000, :boundingbox=>[7, -19, 994, 703]}, 191=>{:wx=>611, :boundingbox=>[91, -201, 527, 525]}, 193=>{:wx=>333, :boundingbox=>[14, 593, 211, 734]}, 194=>{:wx=>333, :boundingbox=>[122, 593, 319, 734]}, 195=>{:wx=>333, :boundingbox=>[21, 593, 312, 734]}, 196=>{:wx=>333, :boundingbox=>[-4, 606, 337, 722]}, 197=>{:wx=>333, :boundingbox=>[10, 627, 323, 684]}, 198=>{:wx=>333, :boundingbox=>[13, 595, 321, 731]}, 199=>{:wx=>333, :boundingbox=>[121, 604, 212, 706]}, 200=>{:wx=>333, :boundingbox=>[40, 604, 293, 706]}, 202=>{:wx=>333, :boundingbox=>[75, 572, 259, 756]}, 203=>{:wx=>333, :boundingbox=>[45, -225, 259, 0]}, 205=>{:wx=>333, :boundingbox=>[31, 593, 409, 734]}, 206=>{:wx=>333, :boundingbox=>[73, -225, 287, 0]}, 207=>{:wx=>333, :boundingbox=>[21, 593, 312, 734]}, 208=>{:wx=>1000, :boundingbox=>[0, 240, 1000, 313]}, 225=>{:wx=>1000, :boundingbox=>[8, 0, 951, 718]}, 227=>{:wx=>370, :boundingbox=>[24, 405, 346, 737]}, 232=>{:wx=>556, :boundingbox=>[-20, 0, 537, 718]}, 233=>{:wx=>778, :boundingbox=>[39, -19, 740, 737]}, 234=>{:wx=>1000, :boundingbox=>[36, -19, 965, 737]}, 235=>{:wx=>365, :boundingbox=>[25, 405, 341, 737]}, 241=>{:wx=>889, :boundingbox=>[36, -15, 847, 538]}, 245=>{:wx=>278, :boundingbox=>[95, 0, 183, 523]}, 248=>{:wx=>222, :boundingbox=>[-20, 0, 242, 718]}, 249=>{:wx=>611, :boundingbox=>[28, -22, 537, 545]}, 250=>{:wx=>944, :boundingbox=>[35, -15, 902, 538]}, 251=>{:wx=>611, :boundingbox=>[67, -15, 571, 728]}, -1=>{:wx=>556, :boundingbox=>[0, 0, 0, 0]}}
			helvetica_bold_metrics = {32=>{:wx=>278, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>333, :boundingbox=>[90, 0, 244, 718]}, 34=>{:wx=>474, :boundingbox=>[98, 447, 376, 718]}, 35=>{:wx=>556, :boundingbox=>[18, 0, 538, 698]}, 36=>{:wx=>556, :boundingbox=>[30, -115, 523, 775]}, 37=>{:wx=>889, :boundingbox=>[28, -19, 861, 710]}, 38=>{:wx=>722, :boundingbox=>[54, -19, 701, 718]}, 39=>{:wx=>278, :boundingbox=>[69, 445, 209, 718]}, 40=>{:wx=>333, :boundingbox=>[35, -208, 314, 734]}, 41=>{:wx=>333, :boundingbox=>[19, -208, 298, 734]}, 42=>{:wx=>389, :boundingbox=>[27, 387, 362, 718]}, 43=>{:wx=>584, :boundingbox=>[40, 0, 544, 506]}, 44=>{:wx=>278, :boundingbox=>[64, -168, 214, 146]}, 45=>{:wx=>333, :boundingbox=>[27, 215, 306, 345]}, 46=>{:wx=>278, :boundingbox=>[64, 0, 214, 146]}, 47=>{:wx=>278, :boundingbox=>[-33, -19, 311, 737]}, 48=>{:wx=>556, :boundingbox=>[32, -19, 524, 710]}, 49=>{:wx=>556, :boundingbox=>[69, 0, 378, 710]}, 50=>{:wx=>556, :boundingbox=>[26, 0, 511, 710]}, 51=>{:wx=>556, :boundingbox=>[27, -19, 516, 710]}, 52=>{:wx=>556, :boundingbox=>[27, 0, 526, 710]}, 53=>{:wx=>556, :boundingbox=>[27, -19, 516, 698]}, 54=>{:wx=>556, :boundingbox=>[31, -19, 520, 710]}, 55=>{:wx=>556, :boundingbox=>[25, 0, 528, 698]}, 56=>{:wx=>556, :boundingbox=>[32, -19, 524, 710]}, 57=>{:wx=>556, :boundingbox=>[30, -19, 522, 710]}, 58=>{:wx=>333, :boundingbox=>[92, 0, 242, 512]}, 59=>{:wx=>333, :boundingbox=>[92, -168, 242, 512]}, 60=>{:wx=>584, :boundingbox=>[38, -8, 546, 514]}, 61=>{:wx=>584, :boundingbox=>[40, 87, 544, 419]}, 62=>{:wx=>584, :boundingbox=>[38, -8, 546, 514]}, 63=>{:wx=>611, :boundingbox=>[60, 0, 556, 727]}, 64=>{:wx=>975, :boundingbox=>[118, -19, 856, 737]}, 65=>{:wx=>722, :boundingbox=>[20, 0, 702, 718]}, 66=>{:wx=>722, :boundingbox=>[76, 0, 669, 718]}, 67=>{:wx=>722, :boundingbox=>[44, -19, 684, 737]}, 68=>{:wx=>722, :boundingbox=>[76, 0, 685, 718]}, 69=>{:wx=>667, :boundingbox=>[76, 0, 621, 718]}, 70=>{:wx=>611, :boundingbox=>[76, 0, 587, 718]}, 71=>{:wx=>778, :boundingbox=>[44, -19, 713, 737]}, 72=>{:wx=>722, :boundingbox=>[71, 0, 651, 718]}, 73=>{:wx=>278, :boundingbox=>[64, 0, 214, 718]}, 74=>{:wx=>556, :boundingbox=>[22, -18, 484, 718]}, 75=>{:wx=>722, :boundingbox=>[87, 0, 722, 718]}, 76=>{:wx=>611, :boundingbox=>[76, 0, 583, 718]}, 77=>{:wx=>833, :boundingbox=>[69, 0, 765, 718]}, 78=>{:wx=>722, :boundingbox=>[69, 0, 654, 718]}, 79=>{:wx=>778, :boundingbox=>[44, -19, 734, 737]}, 80=>{:wx=>667, :boundingbox=>[76, 0, 627, 718]}, 81=>{:wx=>778, :boundingbox=>[44, -52, 737, 737]}, 82=>{:wx=>722, :boundingbox=>[76, 0, 677, 718]}, 83=>{:wx=>667, :boundingbox=>[39, -19, 629, 737]}, 84=>{:wx=>611, :boundingbox=>[14, 0, 598, 718]}, 85=>{:wx=>722, :boundingbox=>[72, -19, 651, 718]}, 86=>{:wx=>667, :boundingbox=>[19, 0, 648, 718]}, 87=>{:wx=>944, :boundingbox=>[16, 0, 929, 718]}, 88=>{:wx=>667, :boundingbox=>[14, 0, 653, 718]}, 89=>{:wx=>667, :boundingbox=>[15, 0, 653, 718]}, 90=>{:wx=>611, :boundingbox=>[25, 0, 586, 718]}, 91=>{:wx=>333, :boundingbox=>[63, -196, 309, 722]}, 92=>{:wx=>278, :boundingbox=>[-33, -19, 311, 737]}, 93=>{:wx=>333, :boundingbox=>[24, -196, 270, 722]}, 94=>{:wx=>584, :boundingbox=>[62, 323, 522, 698]}, 95=>{:wx=>556, :boundingbox=>[0, -125, 556, -75]}, 96=>{:wx=>278, :boundingbox=>[69, 454, 209, 727]}, 97=>{:wx=>556, :boundingbox=>[29, -14, 527, 546]}, 98=>{:wx=>611, :boundingbox=>[61, -14, 578, 718]}, 99=>{:wx=>556, :boundingbox=>[34, -14, 524, 546]}, 100=>{:wx=>611, :boundingbox=>[34, -14, 551, 718]}, 101=>{:wx=>556, :boundingbox=>[23, -14, 528, 546]}, 102=>{:wx=>333, :boundingbox=>[10, 0, 318, 727]}, 103=>{:wx=>611, :boundingbox=>[40, -217, 553, 546]}, 104=>{:wx=>611, :boundingbox=>[65, 0, 546, 718]}, 105=>{:wx=>278, :boundingbox=>[69, 0, 209, 725]}, 106=>{:wx=>278, :boundingbox=>[3, -214, 209, 725]}, 107=>{:wx=>556, :boundingbox=>[69, 0, 562, 718]}, 108=>{:wx=>278, :boundingbox=>[69, 0, 209, 718]}, 109=>{:wx=>889, :boundingbox=>[64, 0, 826, 546]}, 110=>{:wx=>611, :boundingbox=>[65, 0, 546, 546]}, 111=>{:wx=>611, :boundingbox=>[34, -14, 578, 546]}, 112=>{:wx=>611, :boundingbox=>[62, -207, 578, 546]}, 113=>{:wx=>611, :boundingbox=>[34, -207, 552, 546]}, 114=>{:wx=>389, :boundingbox=>[64, 0, 373, 546]}, 115=>{:wx=>556, :boundingbox=>[30, -14, 519, 546]}, 116=>{:wx=>333, :boundingbox=>[10, -6, 309, 676]}, 117=>{:wx=>611, :boundingbox=>[66, -14, 545, 532]}, 118=>{:wx=>556, :boundingbox=>[13, 0, 543, 532]}, 119=>{:wx=>778, :boundingbox=>[10, 0, 769, 532]}, 120=>{:wx=>556, :boundingbox=>[15, 0, 541, 532]}, 121=>{:wx=>556, :boundingbox=>[10, -214, 539, 532]}, 122=>{:wx=>500, :boundingbox=>[20, 0, 480, 532]}, 123=>{:wx=>389, :boundingbox=>[48, -196, 365, 722]}, 124=>{:wx=>280, :boundingbox=>[84, -225, 196, 775]}, 125=>{:wx=>389, :boundingbox=>[24, -196, 341, 722]}, 126=>{:wx=>584, :boundingbox=>[61, 163, 523, 343]}, 161=>{:wx=>333, :boundingbox=>[90, -186, 244, 532]}, 162=>{:wx=>556, :boundingbox=>[34, -118, 524, 628]}, 163=>{:wx=>556, :boundingbox=>[28, -16, 541, 718]}, 164=>{:wx=>167, :boundingbox=>[-170, -19, 336, 710]}, 165=>{:wx=>556, :boundingbox=>[-9, 0, 565, 698]}, 166=>{:wx=>556, :boundingbox=>[-10, -210, 516, 737]}, 167=>{:wx=>556, :boundingbox=>[34, -184, 522, 727]}, 168=>{:wx=>556, :boundingbox=>[-3, 76, 559, 636]}, 169=>{:wx=>238, :boundingbox=>[70, 447, 168, 718]}, 170=>{:wx=>500, :boundingbox=>[64, 454, 436, 727]}, 171=>{:wx=>556, :boundingbox=>[88, 76, 468, 484]}, 172=>{:wx=>333, :boundingbox=>[83, 76, 250, 484]}, 173=>{:wx=>333, :boundingbox=>[83, 76, 250, 484]}, 174=>{:wx=>611, :boundingbox=>[10, 0, 542, 727]}, 175=>{:wx=>611, :boundingbox=>[10, 0, 542, 727]}, 177=>{:wx=>556, :boundingbox=>[0, 227, 556, 333]}, 178=>{:wx=>556, :boundingbox=>[36, -171, 520, 718]}, 179=>{:wx=>556, :boundingbox=>[36, -171, 520, 718]}, 180=>{:wx=>278, :boundingbox=>[58, 172, 220, 334]}, 182=>{:wx=>556, :boundingbox=>[-8, -191, 539, 700]}, 183=>{:wx=>350, :boundingbox=>[10, 194, 340, 524]}, 184=>{:wx=>278, :boundingbox=>[69, -146, 209, 127]}, 185=>{:wx=>500, :boundingbox=>[64, -146, 436, 127]}, 186=>{:wx=>500, :boundingbox=>[64, 445, 436, 718]}, 187=>{:wx=>556, :boundingbox=>[88, 76, 468, 484]}, 188=>{:wx=>1000, :boundingbox=>[92, 0, 908, 146]}, 189=>{:wx=>1000, :boundingbox=>[-3, -19, 1003, 710]}, 191=>{:wx=>611, :boundingbox=>[55, -195, 551, 532]}, 193=>{:wx=>333, :boundingbox=>[-23, 604, 225, 750]}, 194=>{:wx=>333, :boundingbox=>[108, 604, 356, 750]}, 195=>{:wx=>333, :boundingbox=>[-10, 604, 343, 750]}, 196=>{:wx=>333, :boundingbox=>[-17, 610, 350, 737]}, 197=>{:wx=>333, :boundingbox=>[-6, 604, 339, 678]}, 198=>{:wx=>333, :boundingbox=>[-2, 604, 335, 750]}, 199=>{:wx=>333, :boundingbox=>[104, 614, 230, 729]}, 200=>{:wx=>333, :boundingbox=>[6, 614, 327, 729]}, 202=>{:wx=>333, :boundingbox=>[59, 568, 275, 776]}, 203=>{:wx=>333, :boundingbox=>[6, -228, 245, 0]}, 205=>{:wx=>333, :boundingbox=>[9, 604, 486, 750]}, 206=>{:wx=>333, :boundingbox=>[71, -228, 304, 0]}, 207=>{:wx=>333, :boundingbox=>[-10, 604, 343, 750]}, 208=>{:wx=>1000, :boundingbox=>[0, 227, 1000, 333]}, 225=>{:wx=>1000, :boundingbox=>[5, 0, 954, 718]}, 227=>{:wx=>370, :boundingbox=>[22, 401, 347, 737]}, 232=>{:wx=>611, :boundingbox=>[-20, 0, 583, 718]}, 233=>{:wx=>778, :boundingbox=>[33, -27, 744, 745]}, 234=>{:wx=>1000, :boundingbox=>[37, -19, 961, 737]}, 235=>{:wx=>365, :boundingbox=>[6, 401, 360, 737]}, 241=>{:wx=>889, :boundingbox=>[29, -14, 858, 546]}, 245=>{:wx=>278, :boundingbox=>[69, 0, 209, 532]}, 248=>{:wx=>278, :boundingbox=>[-18, 0, 296, 718]}, 249=>{:wx=>611, :boundingbox=>[22, -29, 589, 560]}, 250=>{:wx=>944, :boundingbox=>[34, -14, 912, 546]}, 251=>{:wx=>611, :boundingbox=>[69, -14, 579, 731]}, -1=>{:wx=>556, :boundingbox=>[0, 0, 0, 0]}}
			helvetica_boldoblique_metrics = {32=>{:wx=>278, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>333, :boundingbox=>[94, 0, 397, 718]}, 34=>{:wx=>474, :boundingbox=>[193, 447, 529, 718]}, 35=>{:wx=>556, :boundingbox=>[60, 0, 644, 698]}, 36=>{:wx=>556, :boundingbox=>[67, -115, 622, 775]}, 37=>{:wx=>889, :boundingbox=>[136, -19, 901, 710]}, 38=>{:wx=>722, :boundingbox=>[89, -19, 732, 718]}, 39=>{:wx=>278, :boundingbox=>[167, 445, 362, 718]}, 40=>{:wx=>333, :boundingbox=>[76, -208, 470, 734]}, 41=>{:wx=>333, :boundingbox=>[-25, -208, 369, 734]}, 42=>{:wx=>389, :boundingbox=>[146, 387, 481, 718]}, 43=>{:wx=>584, :boundingbox=>[82, 0, 610, 506]}, 44=>{:wx=>278, :boundingbox=>[28, -168, 245, 146]}, 45=>{:wx=>333, :boundingbox=>[73, 215, 379, 345]}, 46=>{:wx=>278, :boundingbox=>[64, 0, 245, 146]}, 47=>{:wx=>278, :boundingbox=>[-37, -19, 468, 737]}, 48=>{:wx=>556, :boundingbox=>[86, -19, 617, 710]}, 49=>{:wx=>556, :boundingbox=>[173, 0, 529, 710]}, 50=>{:wx=>556, :boundingbox=>[26, 0, 619, 710]}, 51=>{:wx=>556, :boundingbox=>[65, -19, 608, 710]}, 52=>{:wx=>556, :boundingbox=>[60, 0, 598, 710]}, 53=>{:wx=>556, :boundingbox=>[64, -19, 636, 698]}, 54=>{:wx=>556, :boundingbox=>[85, -19, 619, 710]}, 55=>{:wx=>556, :boundingbox=>[125, 0, 676, 698]}, 56=>{:wx=>556, :boundingbox=>[69, -19, 616, 710]}, 57=>{:wx=>556, :boundingbox=>[78, -19, 615, 710]}, 58=>{:wx=>333, :boundingbox=>[92, 0, 351, 512]}, 59=>{:wx=>333, :boundingbox=>[56, -168, 351, 512]}, 60=>{:wx=>584, :boundingbox=>[82, -8, 655, 514]}, 61=>{:wx=>584, :boundingbox=>[58, 87, 633, 419]}, 62=>{:wx=>584, :boundingbox=>[36, -8, 609, 514]}, 63=>{:wx=>611, :boundingbox=>[165, 0, 671, 727]}, 64=>{:wx=>975, :boundingbox=>[186, -19, 954, 737]}, 65=>{:wx=>722, :boundingbox=>[20, 0, 702, 718]}, 66=>{:wx=>722, :boundingbox=>[76, 0, 764, 718]}, 67=>{:wx=>722, :boundingbox=>[107, -19, 789, 737]}, 68=>{:wx=>722, :boundingbox=>[76, 0, 777, 718]}, 69=>{:wx=>667, :boundingbox=>[76, 0, 757, 718]}, 70=>{:wx=>611, :boundingbox=>[76, 0, 740, 718]}, 71=>{:wx=>778, :boundingbox=>[108, -19, 817, 737]}, 72=>{:wx=>722, :boundingbox=>[71, 0, 804, 718]}, 73=>{:wx=>278, :boundingbox=>[64, 0, 367, 718]}, 74=>{:wx=>556, :boundingbox=>[60, -18, 637, 718]}, 75=>{:wx=>722, :boundingbox=>[87, 0, 858, 718]}, 76=>{:wx=>611, :boundingbox=>[76, 0, 611, 718]}, 77=>{:wx=>833, :boundingbox=>[69, 0, 918, 718]}, 78=>{:wx=>722, :boundingbox=>[69, 0, 807, 718]}, 79=>{:wx=>778, :boundingbox=>[107, -19, 823, 737]}, 80=>{:wx=>667, :boundingbox=>[76, 0, 738, 718]}, 81=>{:wx=>778, :boundingbox=>[107, -52, 823, 737]}, 82=>{:wx=>722, :boundingbox=>[76, 0, 778, 718]}, 83=>{:wx=>667, :boundingbox=>[81, -19, 718, 737]}, 84=>{:wx=>611, :boundingbox=>[140, 0, 751, 718]}, 85=>{:wx=>722, :boundingbox=>[116, -19, 804, 718]}, 86=>{:wx=>667, :boundingbox=>[172, 0, 801, 718]}, 87=>{:wx=>944, :boundingbox=>[169, 0, 1082, 718]}, 88=>{:wx=>667, :boundingbox=>[14, 0, 791, 718]}, 89=>{:wx=>667, :boundingbox=>[168, 0, 806, 718]}, 90=>{:wx=>611, :boundingbox=>[25, 0, 737, 718]}, 91=>{:wx=>333, :boundingbox=>[21, -196, 462, 722]}, 92=>{:wx=>278, :boundingbox=>[124, -19, 307, 737]}, 93=>{:wx=>333, :boundingbox=>[-18, -196, 423, 722]}, 94=>{:wx=>584, :boundingbox=>[131, 323, 591, 698]}, 95=>{:wx=>556, :boundingbox=>[-27, -125, 540, -75]}, 96=>{:wx=>278, :boundingbox=>[165, 454, 361, 727]}, 97=>{:wx=>556, :boundingbox=>[55, -14, 583, 546]}, 98=>{:wx=>611, :boundingbox=>[61, -14, 645, 718]}, 99=>{:wx=>556, :boundingbox=>[79, -14, 599, 546]}, 100=>{:wx=>611, :boundingbox=>[82, -14, 704, 718]}, 101=>{:wx=>556, :boundingbox=>[70, -14, 593, 546]}, 102=>{:wx=>333, :boundingbox=>[87, 0, 469, 727]}, 103=>{:wx=>611, :boundingbox=>[38, -217, 666, 546]}, 104=>{:wx=>611, :boundingbox=>[65, 0, 629, 718]}, 105=>{:wx=>278, :boundingbox=>[69, 0, 363, 725]}, 106=>{:wx=>278, :boundingbox=>[-42, -214, 363, 725]}, 107=>{:wx=>556, :boundingbox=>[69, 0, 670, 718]}, 108=>{:wx=>278, :boundingbox=>[69, 0, 362, 718]}, 109=>{:wx=>889, :boundingbox=>[64, 0, 909, 546]}, 110=>{:wx=>611, :boundingbox=>[65, 0, 629, 546]}, 111=>{:wx=>611, :boundingbox=>[82, -14, 643, 546]}, 112=>{:wx=>611, :boundingbox=>[18, -207, 645, 546]}, 113=>{:wx=>611, :boundingbox=>[80, -207, 665, 546]}, 114=>{:wx=>389, :boundingbox=>[64, 0, 489, 546]}, 115=>{:wx=>556, :boundingbox=>[63, -14, 584, 546]}, 116=>{:wx=>333, :boundingbox=>[100, -6, 422, 676]}, 117=>{:wx=>611, :boundingbox=>[98, -14, 658, 532]}, 118=>{:wx=>556, :boundingbox=>[126, 0, 656, 532]}, 119=>{:wx=>778, :boundingbox=>[123, 0, 882, 532]}, 120=>{:wx=>556, :boundingbox=>[15, 0, 648, 532]}, 121=>{:wx=>556, :boundingbox=>[42, -214, 652, 532]}, 122=>{:wx=>500, :boundingbox=>[20, 0, 583, 532]}, 123=>{:wx=>389, :boundingbox=>[94, -196, 518, 722]}, 124=>{:wx=>280, :boundingbox=>[36, -225, 361, 775]}, 125=>{:wx=>389, :boundingbox=>[-18, -196, 407, 722]}, 126=>{:wx=>584, :boundingbox=>[115, 163, 577, 343]}, 161=>{:wx=>333, :boundingbox=>[50, -186, 353, 532]}, 162=>{:wx=>556, :boundingbox=>[79, -118, 599, 628]}, 163=>{:wx=>556, :boundingbox=>[50, -16, 635, 718]}, 164=>{:wx=>167, :boundingbox=>[-174, -19, 487, 710]}, 165=>{:wx=>556, :boundingbox=>[60, 0, 713, 698]}, 166=>{:wx=>556, :boundingbox=>[-50, -210, 669, 737]}, 167=>{:wx=>556, :boundingbox=>[61, -184, 598, 727]}, 168=>{:wx=>556, :boundingbox=>[27, 76, 680, 636]}, 169=>{:wx=>238, :boundingbox=>[165, 447, 321, 718]}, 170=>{:wx=>500, :boundingbox=>[160, 454, 588, 727]}, 171=>{:wx=>556, :boundingbox=>[135, 76, 571, 484]}, 172=>{:wx=>333, :boundingbox=>[130, 76, 353, 484]}, 173=>{:wx=>333, :boundingbox=>[99, 76, 322, 484]}, 174=>{:wx=>611, :boundingbox=>[87, 0, 696, 727]}, 175=>{:wx=>611, :boundingbox=>[87, 0, 695, 727]}, 177=>{:wx=>556, :boundingbox=>[48, 227, 627, 333]}, 178=>{:wx=>556, :boundingbox=>[118, -171, 626, 718]}, 179=>{:wx=>556, :boundingbox=>[46, -171, 628, 718]}, 180=>{:wx=>278, :boundingbox=>[110, 172, 276, 334]}, 182=>{:wx=>556, :boundingbox=>[98, -191, 688, 700]}, 183=>{:wx=>350, :boundingbox=>[83, 194, 420, 524]}, 184=>{:wx=>278, :boundingbox=>[41, -146, 236, 127]}, 185=>{:wx=>500, :boundingbox=>[36, -146, 463, 127]}, 186=>{:wx=>500, :boundingbox=>[162, 445, 589, 718]}, 187=>{:wx=>556, :boundingbox=>[104, 76, 540, 484]}, 188=>{:wx=>1000, :boundingbox=>[92, 0, 939, 146]}, 189=>{:wx=>1000, :boundingbox=>[76, -19, 1038, 710]}, 191=>{:wx=>611, :boundingbox=>[53, -195, 559, 532]}, 193=>{:wx=>333, :boundingbox=>[136, 604, 353, 750]}, 194=>{:wx=>333, :boundingbox=>[236, 604, 515, 750]}, 195=>{:wx=>333, :boundingbox=>[118, 604, 471, 750]}, 196=>{:wx=>333, :boundingbox=>[113, 610, 507, 737]}, 197=>{:wx=>333, :boundingbox=>[122, 604, 483, 678]}, 198=>{:wx=>333, :boundingbox=>[156, 604, 494, 750]}, 199=>{:wx=>333, :boundingbox=>[235, 614, 385, 729]}, 200=>{:wx=>333, :boundingbox=>[137, 614, 482, 729]}, 202=>{:wx=>333, :boundingbox=>[200, 568, 420, 776]}, 203=>{:wx=>333, :boundingbox=>[-37, -228, 220, 0]}, 205=>{:wx=>333, :boundingbox=>[137, 604, 645, 750]}, 206=>{:wx=>333, :boundingbox=>[41, -228, 264, 0]}, 207=>{:wx=>333, :boundingbox=>[149, 604, 502, 750]}, 208=>{:wx=>1000, :boundingbox=>[48, 227, 1071, 333]}, 225=>{:wx=>1000, :boundingbox=>[5, 0, 1100, 718]}, 227=>{:wx=>370, :boundingbox=>[125, 401, 465, 737]}, 232=>{:wx=>611, :boundingbox=>[34, 0, 611, 718]}, 233=>{:wx=>778, :boundingbox=>[35, -27, 894, 745]}, 234=>{:wx=>1000, :boundingbox=>[99, -19, 1114, 737]}, 235=>{:wx=>365, :boundingbox=>[123, 401, 485, 737]}, 241=>{:wx=>889, :boundingbox=>[56, -14, 923, 546]}, 245=>{:wx=>278, :boundingbox=>[69, 0, 322, 532]}, 248=>{:wx=>278, :boundingbox=>[40, 0, 407, 718]}, 249=>{:wx=>611, :boundingbox=>[22, -29, 701, 560]}, 250=>{:wx=>944, :boundingbox=>[82, -14, 977, 546]}, 251=>{:wx=>611, :boundingbox=>[69, -14, 657, 731]}, -1=>{:wx=>556, :boundingbox=>[0, 0, 0, 0]}}
			helvetica_oblique_metrics = {32=>{:wx=>278, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>278, :boundingbox=>[90, 0, 340, 718]}, 34=>{:wx=>355, :boundingbox=>[168, 463, 438, 718]}, 35=>{:wx=>556, :boundingbox=>[73, 0, 631, 688]}, 36=>{:wx=>556, :boundingbox=>[69, -115, 617, 775]}, 37=>{:wx=>889, :boundingbox=>[147, -19, 889, 703]}, 38=>{:wx=>667, :boundingbox=>[77, -15, 647, 718]}, 39=>{:wx=>222, :boundingbox=>[151, 463, 310, 718]}, 40=>{:wx=>333, :boundingbox=>[108, -207, 454, 733]}, 41=>{:wx=>333, :boundingbox=>[-9, -207, 337, 733]}, 42=>{:wx=>389, :boundingbox=>[165, 431, 475, 718]}, 43=>{:wx=>584, :boundingbox=>[85, 0, 606, 505]}, 44=>{:wx=>278, :boundingbox=>[56, -147, 214, 106]}, 45=>{:wx=>333, :boundingbox=>[93, 232, 357, 322]}, 46=>{:wx=>278, :boundingbox=>[87, 0, 214, 106]}, 47=>{:wx=>278, :boundingbox=>[-21, -19, 452, 737]}, 48=>{:wx=>556, :boundingbox=>[93, -19, 608, 703]}, 49=>{:wx=>556, :boundingbox=>[207, 0, 508, 703]}, 50=>{:wx=>556, :boundingbox=>[26, 0, 617, 703]}, 51=>{:wx=>556, :boundingbox=>[75, -19, 610, 703]}, 52=>{:wx=>556, :boundingbox=>[61, 0, 576, 703]}, 53=>{:wx=>556, :boundingbox=>[68, -19, 621, 688]}, 54=>{:wx=>556, :boundingbox=>[91, -19, 615, 703]}, 55=>{:wx=>556, :boundingbox=>[137, 0, 669, 688]}, 56=>{:wx=>556, :boundingbox=>[74, -19, 607, 703]}, 57=>{:wx=>556, :boundingbox=>[82, -19, 609, 703]}, 58=>{:wx=>278, :boundingbox=>[87, 0, 301, 516]}, 59=>{:wx=>278, :boundingbox=>[56, -147, 301, 516]}, 60=>{:wx=>584, :boundingbox=>[94, 11, 641, 495]}, 61=>{:wx=>584, :boundingbox=>[63, 115, 628, 390]}, 62=>{:wx=>584, :boundingbox=>[50, 11, 597, 495]}, 63=>{:wx=>556, :boundingbox=>[161, 0, 610, 727]}, 64=>{:wx=>1015, :boundingbox=>[215, -19, 965, 737]}, 65=>{:wx=>667, :boundingbox=>[14, 0, 654, 718]}, 66=>{:wx=>667, :boundingbox=>[74, 0, 712, 718]}, 67=>{:wx=>722, :boundingbox=>[108, -19, 782, 737]}, 68=>{:wx=>722, :boundingbox=>[81, 0, 764, 718]}, 69=>{:wx=>667, :boundingbox=>[86, 0, 762, 718]}, 70=>{:wx=>611, :boundingbox=>[86, 0, 736, 718]}, 71=>{:wx=>778, :boundingbox=>[111, -19, 799, 737]}, 72=>{:wx=>722, :boundingbox=>[77, 0, 799, 718]}, 73=>{:wx=>278, :boundingbox=>[91, 0, 341, 718]}, 74=>{:wx=>500, :boundingbox=>[47, -19, 581, 718]}, 75=>{:wx=>667, :boundingbox=>[76, 0, 808, 718]}, 76=>{:wx=>556, :boundingbox=>[76, 0, 555, 718]}, 77=>{:wx=>833, :boundingbox=>[73, 0, 914, 718]}, 78=>{:wx=>722, :boundingbox=>[76, 0, 799, 718]}, 79=>{:wx=>778, :boundingbox=>[105, -19, 826, 737]}, 80=>{:wx=>667, :boundingbox=>[86, 0, 737, 718]}, 81=>{:wx=>778, :boundingbox=>[105, -56, 826, 737]}, 82=>{:wx=>722, :boundingbox=>[88, 0, 773, 718]}, 83=>{:wx=>667, :boundingbox=>[90, -19, 713, 737]}, 84=>{:wx=>611, :boundingbox=>[148, 0, 750, 718]}, 85=>{:wx=>722, :boundingbox=>[123, -19, 797, 718]}, 86=>{:wx=>667, :boundingbox=>[173, 0, 800, 718]}, 87=>{:wx=>944, :boundingbox=>[169, 0, 1081, 718]}, 88=>{:wx=>667, :boundingbox=>[19, 0, 790, 718]}, 89=>{:wx=>667, :boundingbox=>[167, 0, 806, 718]}, 90=>{:wx=>611, :boundingbox=>[23, 0, 741, 718]}, 91=>{:wx=>278, :boundingbox=>[21, -196, 403, 722]}, 92=>{:wx=>278, :boundingbox=>[140, -19, 291, 737]}, 93=>{:wx=>278, :boundingbox=>[-14, -196, 368, 722]}, 94=>{:wx=>469, :boundingbox=>[42, 264, 539, 688]}, 95=>{:wx=>556, :boundingbox=>[-27, -125, 540, -75]}, 96=>{:wx=>222, :boundingbox=>[165, 470, 323, 725]}, 97=>{:wx=>556, :boundingbox=>[61, -15, 559, 538]}, 98=>{:wx=>556, :boundingbox=>[58, -15, 584, 718]}, 99=>{:wx=>500, :boundingbox=>[74, -15, 553, 538]}, 100=>{:wx=>556, :boundingbox=>[84, -15, 652, 718]}, 101=>{:wx=>556, :boundingbox=>[84, -15, 578, 538]}, 102=>{:wx=>278, :boundingbox=>[86, 0, 416, 728]}, 103=>{:wx=>556, :boundingbox=>[42, -220, 610, 538]}, 104=>{:wx=>556, :boundingbox=>[65, 0, 573, 718]}, 105=>{:wx=>222, :boundingbox=>[67, 0, 308, 718]}, 106=>{:wx=>222, :boundingbox=>[-60, -210, 308, 718]}, 107=>{:wx=>500, :boundingbox=>[67, 0, 600, 718]}, 108=>{:wx=>222, :boundingbox=>[67, 0, 308, 718]}, 109=>{:wx=>833, :boundingbox=>[65, 0, 852, 538]}, 110=>{:wx=>556, :boundingbox=>[65, 0, 573, 538]}, 111=>{:wx=>556, :boundingbox=>[83, -14, 585, 538]}, 112=>{:wx=>556, :boundingbox=>[14, -207, 584, 538]}, 113=>{:wx=>556, :boundingbox=>[84, -207, 605, 538]}, 114=>{:wx=>333, :boundingbox=>[77, 0, 446, 538]}, 115=>{:wx=>500, :boundingbox=>[63, -15, 529, 538]}, 116=>{:wx=>278, :boundingbox=>[102, -7, 368, 669]}, 117=>{:wx=>556, :boundingbox=>[94, -15, 600, 523]}, 118=>{:wx=>500, :boundingbox=>[119, 0, 603, 523]}, 119=>{:wx=>722, :boundingbox=>[125, 0, 820, 523]}, 120=>{:wx=>500, :boundingbox=>[11, 0, 594, 523]}, 121=>{:wx=>500, :boundingbox=>[15, -214, 600, 523]}, 122=>{:wx=>500, :boundingbox=>[31, 0, 571, 523]}, 123=>{:wx=>334, :boundingbox=>[92, -196, 445, 722]}, 124=>{:wx=>260, :boundingbox=>[46, -225, 332, 775]}, 125=>{:wx=>334, :boundingbox=>[0, -196, 354, 722]}, 126=>{:wx=>584, :boundingbox=>[111, 180, 580, 326]}, 161=>{:wx=>333, :boundingbox=>[77, -195, 326, 523]}, 162=>{:wx=>556, :boundingbox=>[95, -115, 584, 623]}, 163=>{:wx=>556, :boundingbox=>[49, -16, 634, 718]}, 164=>{:wx=>167, :boundingbox=>[-170, -19, 482, 703]}, 165=>{:wx=>556, :boundingbox=>[81, 0, 699, 688]}, 166=>{:wx=>556, :boundingbox=>[-52, -207, 654, 737]}, 167=>{:wx=>556, :boundingbox=>[76, -191, 584, 737]}, 168=>{:wx=>556, :boundingbox=>[60, 99, 646, 603]}, 169=>{:wx=>191, :boundingbox=>[157, 463, 285, 718]}, 170=>{:wx=>333, :boundingbox=>[138, 470, 461, 725]}, 171=>{:wx=>556, :boundingbox=>[146, 108, 554, 446]}, 172=>{:wx=>333, :boundingbox=>[137, 108, 340, 446]}, 173=>{:wx=>333, :boundingbox=>[111, 108, 314, 446]}, 174=>{:wx=>500, :boundingbox=>[86, 0, 587, 728]}, 175=>{:wx=>500, :boundingbox=>[86, 0, 585, 728]}, 177=>{:wx=>556, :boundingbox=>[51, 240, 623, 313]}, 178=>{:wx=>556, :boundingbox=>[135, -159, 622, 718]}, 179=>{:wx=>556, :boundingbox=>[52, -159, 623, 718]}, 180=>{:wx=>278, :boundingbox=>[129, 190, 257, 315]}, 182=>{:wx=>537, :boundingbox=>[126, -173, 650, 718]}, 183=>{:wx=>350, :boundingbox=>[91, 202, 413, 517]}, 184=>{:wx=>222, :boundingbox=>[21, -149, 180, 106]}, 185=>{:wx=>333, :boundingbox=>[-6, -149, 318, 106]}, 186=>{:wx=>333, :boundingbox=>[124, 463, 448, 718]}, 187=>{:wx=>556, :boundingbox=>[120, 108, 528, 446]}, 188=>{:wx=>1000, :boundingbox=>[115, 0, 908, 106]}, 189=>{:wx=>1000, :boundingbox=>[88, -19, 1029, 703]}, 191=>{:wx=>611, :boundingbox=>[85, -201, 534, 525]}, 193=>{:wx=>333, :boundingbox=>[170, 593, 337, 734]}, 194=>{:wx=>333, :boundingbox=>[248, 593, 475, 734]}, 195=>{:wx=>333, :boundingbox=>[147, 593, 438, 734]}, 196=>{:wx=>333, :boundingbox=>[125, 606, 490, 722]}, 197=>{:wx=>333, :boundingbox=>[143, 627, 468, 684]}, 198=>{:wx=>333, :boundingbox=>[167, 595, 476, 731]}, 199=>{:wx=>333, :boundingbox=>[249, 604, 362, 706]}, 200=>{:wx=>333, :boundingbox=>[168, 604, 443, 706]}, 202=>{:wx=>333, :boundingbox=>[214, 572, 402, 756]}, 203=>{:wx=>333, :boundingbox=>[2, -225, 232, 0]}, 205=>{:wx=>333, :boundingbox=>[157, 593, 565, 734]}, 206=>{:wx=>333, :boundingbox=>[43, -225, 249, 0]}, 207=>{:wx=>333, :boundingbox=>[177, 593, 468, 734]}, 208=>{:wx=>1000, :boundingbox=>[51, 240, 1067, 313]}, 225=>{:wx=>1000, :boundingbox=>[8, 0, 1097, 718]}, 227=>{:wx=>370, :boundingbox=>[127, 405, 449, 737]}, 232=>{:wx=>556, :boundingbox=>[41, 0, 555, 718]}, 233=>{:wx=>778, :boundingbox=>[43, -19, 890, 737]}, 234=>{:wx=>1000, :boundingbox=>[98, -19, 1116, 737]}, 235=>{:wx=>365, :boundingbox=>[141, 405, 468, 737]}, 241=>{:wx=>889, :boundingbox=>[61, -15, 909, 538]}, 245=>{:wx=>278, :boundingbox=>[95, 0, 294, 523]}, 248=>{:wx=>222, :boundingbox=>[41, 0, 347, 718]}, 249=>{:wx=>611, :boundingbox=>[29, -22, 647, 545]}, 250=>{:wx=>944, :boundingbox=>[83, -15, 964, 538]}, 251=>{:wx=>611, :boundingbox=>[67, -15, 658, 728]}, -1=>{:wx=>556, :boundingbox=>[0, 0, 0, 0]}}
			courier_metrics = {32=>{:wx=>600, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>600, :boundingbox=>[236, -15, 364, 572]}, 34=>{:wx=>600, :boundingbox=>[187, 328, 413, 562]}, 35=>{:wx=>600, :boundingbox=>[93, -32, 507, 639]}, 36=>{:wx=>600, :boundingbox=>[105, -126, 496, 662]}, 37=>{:wx=>600, :boundingbox=>[81, -15, 518, 622]}, 38=>{:wx=>600, :boundingbox=>[63, -15, 538, 543]}, 39=>{:wx=>600, :boundingbox=>[213, 328, 376, 562]}, 40=>{:wx=>600, :boundingbox=>[269, -108, 440, 622]}, 41=>{:wx=>600, :boundingbox=>[160, -108, 331, 622]}, 42=>{:wx=>600, :boundingbox=>[116, 257, 484, 607]}, 43=>{:wx=>600, :boundingbox=>[80, 44, 520, 470]}, 44=>{:wx=>600, :boundingbox=>[181, -112, 344, 122]}, 45=>{:wx=>600, :boundingbox=>[103, 231, 497, 285]}, 46=>{:wx=>600, :boundingbox=>[229, -15, 371, 109]}, 47=>{:wx=>600, :boundingbox=>[125, -80, 475, 629]}, 48=>{:wx=>600, :boundingbox=>[106, -15, 494, 622]}, 49=>{:wx=>600, :boundingbox=>[96, 0, 505, 622]}, 50=>{:wx=>600, :boundingbox=>[70, 0, 471, 622]}, 51=>{:wx=>600, :boundingbox=>[75, -15, 466, 622]}, 52=>{:wx=>600, :boundingbox=>[78, 0, 500, 622]}, 53=>{:wx=>600, :boundingbox=>[92, -15, 497, 607]}, 54=>{:wx=>600, :boundingbox=>[111, -15, 497, 622]}, 55=>{:wx=>600, :boundingbox=>[82, 0, 483, 607]}, 56=>{:wx=>600, :boundingbox=>[102, -15, 498, 622]}, 57=>{:wx=>600, :boundingbox=>[96, -15, 489, 622]}, 58=>{:wx=>600, :boundingbox=>[229, -15, 371, 385]}, 59=>{:wx=>600, :boundingbox=>[181, -112, 371, 385]}, 60=>{:wx=>600, :boundingbox=>[41, 42, 519, 472]}, 61=>{:wx=>600, :boundingbox=>[80, 138, 520, 376]}, 62=>{:wx=>600, :boundingbox=>[66, 42, 544, 472]}, 63=>{:wx=>600, :boundingbox=>[129, -15, 492, 572]}, 64=>{:wx=>600, :boundingbox=>[77, -15, 533, 622]}, 65=>{:wx=>600, :boundingbox=>[3, 0, 597, 562]}, 66=>{:wx=>600, :boundingbox=>[43, 0, 559, 562]}, 67=>{:wx=>600, :boundingbox=>[41, -18, 540, 580]}, 68=>{:wx=>600, :boundingbox=>[43, 0, 574, 562]}, 69=>{:wx=>600, :boundingbox=>[53, 0, 550, 562]}, 70=>{:wx=>600, :boundingbox=>[53, 0, 545, 562]}, 71=>{:wx=>600, :boundingbox=>[31, -18, 575, 580]}, 72=>{:wx=>600, :boundingbox=>[32, 0, 568, 562]}, 73=>{:wx=>600, :boundingbox=>[96, 0, 504, 562]}, 74=>{:wx=>600, :boundingbox=>[34, -18, 566, 562]}, 75=>{:wx=>600, :boundingbox=>[38, 0, 582, 562]}, 76=>{:wx=>600, :boundingbox=>[47, 0, 554, 562]}, 77=>{:wx=>600, :boundingbox=>[4, 0, 596, 562]}, 78=>{:wx=>600, :boundingbox=>[7, -13, 593, 562]}, 79=>{:wx=>600, :boundingbox=>[43, -18, 557, 580]}, 80=>{:wx=>600, :boundingbox=>[79, 0, 558, 562]}, 81=>{:wx=>600, :boundingbox=>[43, -138, 557, 580]}, 82=>{:wx=>600, :boundingbox=>[38, 0, 588, 562]}, 83=>{:wx=>600, :boundingbox=>[72, -20, 529, 580]}, 84=>{:wx=>600, :boundingbox=>[38, 0, 563, 562]}, 85=>{:wx=>600, :boundingbox=>[17, -18, 583, 562]}, 86=>{:wx=>600, :boundingbox=>[-4, -13, 604, 562]}, 87=>{:wx=>600, :boundingbox=>[-3, -13, 603, 562]}, 88=>{:wx=>600, :boundingbox=>[23, 0, 577, 562]}, 89=>{:wx=>600, :boundingbox=>[24, 0, 576, 562]}, 90=>{:wx=>600, :boundingbox=>[86, 0, 514, 562]}, 91=>{:wx=>600, :boundingbox=>[269, -108, 442, 622]}, 92=>{:wx=>600, :boundingbox=>[118, -80, 482, 629]}, 93=>{:wx=>600, :boundingbox=>[158, -108, 331, 622]}, 94=>{:wx=>600, :boundingbox=>[94, 354, 506, 622]}, 95=>{:wx=>600, :boundingbox=>[0, -125, 600, -75]}, 96=>{:wx=>600, :boundingbox=>[224, 328, 387, 562]}, 97=>{:wx=>600, :boundingbox=>[53, -15, 559, 441]}, 98=>{:wx=>600, :boundingbox=>[14, -15, 575, 629]}, 99=>{:wx=>600, :boundingbox=>[66, -15, 529, 441]}, 100=>{:wx=>600, :boundingbox=>[45, -15, 591, 629]}, 101=>{:wx=>600, :boundingbox=>[66, -15, 548, 441]}, 102=>{:wx=>600, :boundingbox=>[114, 0, 531, 629]}, 103=>{:wx=>600, :boundingbox=>[45, -157, 566, 441]}, 104=>{:wx=>600, :boundingbox=>[18, 0, 582, 629]}, 105=>{:wx=>600, :boundingbox=>[95, 0, 505, 657]}, 106=>{:wx=>600, :boundingbox=>[82, -157, 410, 657]}, 107=>{:wx=>600, :boundingbox=>[43, 0, 580, 629]}, 108=>{:wx=>600, :boundingbox=>[95, 0, 505, 629]}, 109=>{:wx=>600, :boundingbox=>[-5, 0, 605, 441]}, 110=>{:wx=>600, :boundingbox=>[26, 0, 575, 441]}, 111=>{:wx=>600, :boundingbox=>[62, -15, 538, 441]}, 112=>{:wx=>600, :boundingbox=>[9, -157, 555, 441]}, 113=>{:wx=>600, :boundingbox=>[45, -157, 591, 441]}, 114=>{:wx=>600, :boundingbox=>[60, 0, 559, 441]}, 115=>{:wx=>600, :boundingbox=>[80, -15, 513, 441]}, 116=>{:wx=>600, :boundingbox=>[87, -15, 530, 561]}, 117=>{:wx=>600, :boundingbox=>[21, -15, 562, 426]}, 118=>{:wx=>600, :boundingbox=>[10, -10, 590, 426]}, 119=>{:wx=>600, :boundingbox=>[-4, -10, 604, 426]}, 120=>{:wx=>600, :boundingbox=>[20, 0, 580, 426]}, 121=>{:wx=>600, :boundingbox=>[7, -157, 592, 426]}, 122=>{:wx=>600, :boundingbox=>[99, 0, 502, 426]}, 123=>{:wx=>600, :boundingbox=>[182, -108, 437, 622]}, 124=>{:wx=>600, :boundingbox=>[275, -250, 326, 750]}, 125=>{:wx=>600, :boundingbox=>[163, -108, 418, 622]}, 126=>{:wx=>600, :boundingbox=>[63, 197, 540, 320]}, 161=>{:wx=>600, :boundingbox=>[236, -157, 364, 430]}, 162=>{:wx=>600, :boundingbox=>[96, -49, 500, 614]}, 163=>{:wx=>600, :boundingbox=>[84, -21, 521, 611]}, 164=>{:wx=>600, :boundingbox=>[92, -57, 509, 665]}, 165=>{:wx=>600, :boundingbox=>[26, 0, 574, 562]}, 166=>{:wx=>600, :boundingbox=>[4, -143, 539, 622]}, 167=>{:wx=>600, :boundingbox=>[113, -78, 488, 580]}, 168=>{:wx=>600, :boundingbox=>[73, 58, 527, 506]}, 169=>{:wx=>600, :boundingbox=>[259, 328, 341, 562]}, 170=>{:wx=>600, :boundingbox=>[143, 328, 471, 562]}, 171=>{:wx=>600, :boundingbox=>[37, 70, 563, 446]}, 172=>{:wx=>600, :boundingbox=>[149, 70, 451, 446]}, 173=>{:wx=>600, :boundingbox=>[149, 70, 451, 446]}, 174=>{:wx=>600, :boundingbox=>[3, 0, 597, 629]}, 175=>{:wx=>600, :boundingbox=>[3, 0, 597, 629]}, 177=>{:wx=>600, :boundingbox=>[75, 231, 525, 285]}, 178=>{:wx=>600, :boundingbox=>[141, -78, 459, 580]}, 179=>{:wx=>600, :boundingbox=>[141, -78, 459, 580]}, 180=>{:wx=>600, :boundingbox=>[222, 189, 378, 327]}, 182=>{:wx=>600, :boundingbox=>[50, -78, 511, 562]}, 183=>{:wx=>600, :boundingbox=>[172, 130, 428, 383]}, 184=>{:wx=>600, :boundingbox=>[213, -134, 376, 100]}, 185=>{:wx=>600, :boundingbox=>[143, -134, 457, 100]}, 186=>{:wx=>600, :boundingbox=>[143, 328, 457, 562]}, 187=>{:wx=>600, :boundingbox=>[37, 70, 563, 446]}, 188=>{:wx=>600, :boundingbox=>[37, -15, 563, 111]}, 189=>{:wx=>600, :boundingbox=>[3, -15, 600, 622]}, 191=>{:wx=>600, :boundingbox=>[108, -157, 471, 430]}, 193=>{:wx=>600, :boundingbox=>[151, 497, 378, 672]}, 194=>{:wx=>600, :boundingbox=>[242, 497, 469, 672]}, 195=>{:wx=>600, :boundingbox=>[124, 477, 476, 654]}, 196=>{:wx=>600, :boundingbox=>[105, 489, 503, 606]}, 197=>{:wx=>600, :boundingbox=>[120, 525, 480, 565]}, 198=>{:wx=>600, :boundingbox=>[153, 501, 447, 609]}, 199=>{:wx=>600, :boundingbox=>[249, 537, 352, 640]}, 200=>{:wx=>600, :boundingbox=>[148, 537, 453, 640]}, 202=>{:wx=>600, :boundingbox=>[218, 463, 382, 627]}, 203=>{:wx=>600, :boundingbox=>[224, -151, 362, 10]}, 205=>{:wx=>600, :boundingbox=>[133, 497, 540, 672]}, 206=>{:wx=>600, :boundingbox=>[211, -172, 407, 4]}, 207=>{:wx=>600, :boundingbox=>[124, 492, 476, 669]}, 208=>{:wx=>600, :boundingbox=>[0, 231, 600, 285]}, 225=>{:wx=>600, :boundingbox=>[3, 0, 550, 562]}, 227=>{:wx=>600, :boundingbox=>[156, 249, 442, 580]}, 232=>{:wx=>600, :boundingbox=>[47, 0, 554, 562]}, 233=>{:wx=>600, :boundingbox=>[43, -80, 557, 629]}, 234=>{:wx=>600, :boundingbox=>[7, 0, 567, 562]}, 235=>{:wx=>600, :boundingbox=>[157, 249, 443, 580]}, 241=>{:wx=>600, :boundingbox=>[19, -15, 570, 441]}, 245=>{:wx=>600, :boundingbox=>[95, 0, 505, 426]}, 248=>{:wx=>600, :boundingbox=>[95, 0, 505, 629]}, 249=>{:wx=>600, :boundingbox=>[62, -80, 538, 506]}, 250=>{:wx=>600, :boundingbox=>[19, -15, 559, 441]}, 251=>{:wx=>600, :boundingbox=>[48, -15, 588, 629]}, -1=>{:wx=>600, :boundingbox=>[0, 0, 0, 0]}}
			courier_bold_metrics = {32=>{:wx=>600, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>600, :boundingbox=>[202, -15, 398, 572]}, 34=>{:wx=>600, :boundingbox=>[135, 277, 465, 562]}, 35=>{:wx=>600, :boundingbox=>[56, -45, 544, 651]}, 36=>{:wx=>600, :boundingbox=>[82, -126, 519, 666]}, 37=>{:wx=>600, :boundingbox=>[5, -15, 595, 616]}, 38=>{:wx=>600, :boundingbox=>[36, -15, 546, 543]}, 39=>{:wx=>600, :boundingbox=>[171, 277, 423, 562]}, 40=>{:wx=>600, :boundingbox=>[219, -102, 461, 616]}, 41=>{:wx=>600, :boundingbox=>[139, -102, 381, 616]}, 42=>{:wx=>600, :boundingbox=>[91, 219, 509, 601]}, 43=>{:wx=>600, :boundingbox=>[71, 39, 529, 478]}, 44=>{:wx=>600, :boundingbox=>[123, -111, 393, 174]}, 45=>{:wx=>600, :boundingbox=>[100, 203, 500, 313]}, 46=>{:wx=>600, :boundingbox=>[192, -15, 408, 171]}, 47=>{:wx=>600, :boundingbox=>[98, -77, 502, 626]}, 48=>{:wx=>600, :boundingbox=>[87, -15, 513, 616]}, 49=>{:wx=>600, :boundingbox=>[81, 0, 539, 616]}, 50=>{:wx=>600, :boundingbox=>[61, 0, 499, 616]}, 51=>{:wx=>600, :boundingbox=>[63, -15, 501, 616]}, 52=>{:wx=>600, :boundingbox=>[53, 0, 507, 616]}, 53=>{:wx=>600, :boundingbox=>[70, -15, 521, 601]}, 54=>{:wx=>600, :boundingbox=>[90, -15, 521, 616]}, 55=>{:wx=>600, :boundingbox=>[55, 0, 494, 601]}, 56=>{:wx=>600, :boundingbox=>[83, -15, 517, 616]}, 57=>{:wx=>600, :boundingbox=>[79, -15, 510, 616]}, 58=>{:wx=>600, :boundingbox=>[191, -15, 407, 425]}, 59=>{:wx=>600, :boundingbox=>[123, -111, 408, 425]}, 60=>{:wx=>600, :boundingbox=>[66, 15, 523, 501]}, 61=>{:wx=>600, :boundingbox=>[71, 118, 529, 398]}, 62=>{:wx=>600, :boundingbox=>[77, 15, 534, 501]}, 63=>{:wx=>600, :boundingbox=>[98, -14, 501, 580]}, 64=>{:wx=>600, :boundingbox=>[16, -15, 584, 616]}, 65=>{:wx=>600, :boundingbox=>[-9, 0, 609, 562]}, 66=>{:wx=>600, :boundingbox=>[30, 0, 573, 562]}, 67=>{:wx=>600, :boundingbox=>[22, -18, 560, 580]}, 68=>{:wx=>600, :boundingbox=>[30, 0, 594, 562]}, 69=>{:wx=>600, :boundingbox=>[25, 0, 560, 562]}, 70=>{:wx=>600, :boundingbox=>[39, 0, 570, 562]}, 71=>{:wx=>600, :boundingbox=>[22, -18, 594, 580]}, 72=>{:wx=>600, :boundingbox=>[20, 0, 580, 562]}, 73=>{:wx=>600, :boundingbox=>[77, 0, 523, 562]}, 74=>{:wx=>600, :boundingbox=>[37, -18, 601, 562]}, 75=>{:wx=>600, :boundingbox=>[21, 0, 599, 562]}, 76=>{:wx=>600, :boundingbox=>[39, 0, 578, 562]}, 77=>{:wx=>600, :boundingbox=>[-2, 0, 602, 562]}, 78=>{:wx=>600, :boundingbox=>[8, -12, 610, 562]}, 79=>{:wx=>600, :boundingbox=>[22, -18, 578, 580]}, 80=>{:wx=>600, :boundingbox=>[48, 0, 559, 562]}, 81=>{:wx=>600, :boundingbox=>[32, -138, 578, 580]}, 82=>{:wx=>600, :boundingbox=>[24, 0, 599, 562]}, 83=>{:wx=>600, :boundingbox=>[47, -22, 553, 582]}, 84=>{:wx=>600, :boundingbox=>[21, 0, 579, 562]}, 85=>{:wx=>600, :boundingbox=>[4, -18, 596, 562]}, 86=>{:wx=>600, :boundingbox=>[-13, 0, 613, 562]}, 87=>{:wx=>600, :boundingbox=>[-18, 0, 618, 562]}, 88=>{:wx=>600, :boundingbox=>[12, 0, 588, 562]}, 89=>{:wx=>600, :boundingbox=>[12, 0, 589, 562]}, 90=>{:wx=>600, :boundingbox=>[62, 0, 539, 562]}, 91=>{:wx=>600, :boundingbox=>[245, -102, 475, 616]}, 92=>{:wx=>600, :boundingbox=>[99, -77, 503, 626]}, 93=>{:wx=>600, :boundingbox=>[125, -102, 355, 616]}, 94=>{:wx=>600, :boundingbox=>[108, 250, 492, 616]}, 95=>{:wx=>600, :boundingbox=>[0, -125, 600, -75]}, 96=>{:wx=>600, :boundingbox=>[178, 277, 428, 562]}, 97=>{:wx=>600, :boundingbox=>[35, -15, 570, 454]}, 98=>{:wx=>600, :boundingbox=>[0, -15, 584, 626]}, 99=>{:wx=>600, :boundingbox=>[40, -15, 545, 459]}, 100=>{:wx=>600, :boundingbox=>[20, -15, 591, 626]}, 101=>{:wx=>600, :boundingbox=>[40, -15, 563, 454]}, 102=>{:wx=>600, :boundingbox=>[83, 0, 547, 626]}, 103=>{:wx=>600, :boundingbox=>[30, -146, 580, 454]}, 104=>{:wx=>600, :boundingbox=>[5, 0, 592, 626]}, 105=>{:wx=>600, :boundingbox=>[77, 0, 523, 658]}, 106=>{:wx=>600, :boundingbox=>[63, -146, 440, 658]}, 107=>{:wx=>600, :boundingbox=>[20, 0, 585, 626]}, 108=>{:wx=>600, :boundingbox=>[77, 0, 523, 626]}, 109=>{:wx=>600, :boundingbox=>[-22, 0, 626, 454]}, 110=>{:wx=>600, :boundingbox=>[18, 0, 592, 454]}, 111=>{:wx=>600, :boundingbox=>[30, -15, 570, 454]}, 112=>{:wx=>600, :boundingbox=>[-1, -142, 570, 454]}, 113=>{:wx=>600, :boundingbox=>[20, -142, 591, 454]}, 114=>{:wx=>600, :boundingbox=>[47, 0, 580, 454]}, 115=>{:wx=>600, :boundingbox=>[68, -17, 535, 459]}, 116=>{:wx=>600, :boundingbox=>[47, -15, 532, 562]}, 117=>{:wx=>600, :boundingbox=>[-1, -15, 569, 439]}, 118=>{:wx=>600, :boundingbox=>[-1, 0, 601, 439]}, 119=>{:wx=>600, :boundingbox=>[-18, 0, 618, 439]}, 120=>{:wx=>600, :boundingbox=>[6, 0, 594, 439]}, 121=>{:wx=>600, :boundingbox=>[-4, -142, 601, 439]}, 122=>{:wx=>600, :boundingbox=>[81, 0, 520, 439]}, 123=>{:wx=>600, :boundingbox=>[160, -102, 464, 616]}, 124=>{:wx=>600, :boundingbox=>[255, -250, 345, 750]}, 125=>{:wx=>600, :boundingbox=>[136, -102, 440, 616]}, 126=>{:wx=>600, :boundingbox=>[71, 153, 530, 356]}, 161=>{:wx=>600, :boundingbox=>[202, -146, 398, 449]}, 162=>{:wx=>600, :boundingbox=>[66, -49, 518, 614]}, 163=>{:wx=>600, :boundingbox=>[72, -28, 558, 611]}, 164=>{:wx=>600, :boundingbox=>[25, -60, 576, 661]}, 165=>{:wx=>600, :boundingbox=>[10, 0, 590, 562]}, 166=>{:wx=>600, :boundingbox=>[-30, -131, 572, 616]}, 167=>{:wx=>600, :boundingbox=>[83, -70, 517, 580]}, 168=>{:wx=>600, :boundingbox=>[54, 49, 546, 517]}, 169=>{:wx=>600, :boundingbox=>[227, 277, 373, 562]}, 170=>{:wx=>600, :boundingbox=>[71, 277, 535, 562]}, 171=>{:wx=>600, :boundingbox=>[8, 70, 553, 446]}, 172=>{:wx=>600, :boundingbox=>[141, 70, 459, 446]}, 173=>{:wx=>600, :boundingbox=>[141, 70, 459, 446]}, 174=>{:wx=>600, :boundingbox=>[12, 0, 593, 626]}, 175=>{:wx=>600, :boundingbox=>[12, 0, 593, 626]}, 177=>{:wx=>600, :boundingbox=>[65, 203, 535, 313]}, 178=>{:wx=>600, :boundingbox=>[106, -70, 494, 580]}, 179=>{:wx=>600, :boundingbox=>[106, -70, 494, 580]}, 180=>{:wx=>600, :boundingbox=>[196, 165, 404, 351]}, 182=>{:wx=>600, :boundingbox=>[6, -70, 576, 580]}, 183=>{:wx=>600, :boundingbox=>[140, 132, 460, 430]}, 184=>{:wx=>600, :boundingbox=>[175, -142, 427, 143]}, 185=>{:wx=>600, :boundingbox=>[65, -142, 529, 143]}, 186=>{:wx=>600, :boundingbox=>[61, 277, 525, 562]}, 187=>{:wx=>600, :boundingbox=>[47, 70, 592, 446]}, 188=>{:wx=>600, :boundingbox=>[26, -15, 574, 116]}, 189=>{:wx=>600, :boundingbox=>[-113, -15, 713, 616]}, 191=>{:wx=>600, :boundingbox=>[99, -146, 502, 449]}, 193=>{:wx=>600, :boundingbox=>[132, 508, 395, 661]}, 194=>{:wx=>600, :boundingbox=>[205, 508, 468, 661]}, 195=>{:wx=>600, :boundingbox=>[103, 483, 497, 657]}, 196=>{:wx=>600, :boundingbox=>[89, 493, 512, 636]}, 197=>{:wx=>600, :boundingbox=>[88, 505, 512, 585]}, 198=>{:wx=>600, :boundingbox=>[83, 468, 517, 631]}, 199=>{:wx=>600, :boundingbox=>[230, 498, 370, 638]}, 200=>{:wx=>600, :boundingbox=>[128, 498, 472, 638]}, 202=>{:wx=>600, :boundingbox=>[198, 481, 402, 678]}, 203=>{:wx=>600, :boundingbox=>[205, -206, 387, 0]}, 205=>{:wx=>600, :boundingbox=>[68, 488, 588, 661]}, 206=>{:wx=>600, :boundingbox=>[169, -199, 400, 0]}, 207=>{:wx=>600, :boundingbox=>[103, 493, 497, 667]}, 208=>{:wx=>600, :boundingbox=>[-10, 203, 610, 313]}, 225=>{:wx=>600, :boundingbox=>[-29, 0, 602, 562]}, 227=>{:wx=>600, :boundingbox=>[147, 196, 453, 580]}, 232=>{:wx=>600, :boundingbox=>[39, 0, 578, 562]}, 233=>{:wx=>600, :boundingbox=>[22, -22, 578, 584]}, 234=>{:wx=>600, :boundingbox=>[-25, 0, 595, 562]}, 235=>{:wx=>600, :boundingbox=>[147, 196, 453, 580]}, 241=>{:wx=>600, :boundingbox=>[-4, -15, 601, 454]}, 245=>{:wx=>600, :boundingbox=>[77, 0, 523, 439]}, 248=>{:wx=>600, :boundingbox=>[77, 0, 523, 626]}, 249=>{:wx=>600, :boundingbox=>[30, -24, 570, 463]}, 250=>{:wx=>600, :boundingbox=>[-18, -15, 611, 454]}, 251=>{:wx=>600, :boundingbox=>[22, -15, 596, 626]}, -1=>{:wx=>600, :boundingbox=>[0, 0, 0, 0]}}
			courier_oblique_metrics = {32=>{:wx=>600, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>600, :boundingbox=>[243, -15, 464, 572]}, 34=>{:wx=>600, :boundingbox=>[273, 328, 532, 562]}, 35=>{:wx=>600, :boundingbox=>[133, -32, 596, 639]}, 36=>{:wx=>600, :boundingbox=>[108, -126, 596, 662]}, 37=>{:wx=>600, :boundingbox=>[134, -15, 599, 622]}, 38=>{:wx=>600, :boundingbox=>[87, -15, 580, 543]}, 39=>{:wx=>600, :boundingbox=>[283, 328, 495, 562]}, 40=>{:wx=>600, :boundingbox=>[313, -108, 572, 622]}, 41=>{:wx=>600, :boundingbox=>[137, -108, 396, 622]}, 42=>{:wx=>600, :boundingbox=>[212, 257, 580, 607]}, 43=>{:wx=>600, :boundingbox=>[129, 44, 580, 470]}, 44=>{:wx=>600, :boundingbox=>[157, -112, 370, 122]}, 45=>{:wx=>600, :boundingbox=>[152, 231, 558, 285]}, 46=>{:wx=>600, :boundingbox=>[238, -15, 382, 109]}, 47=>{:wx=>600, :boundingbox=>[112, -80, 604, 629]}, 48=>{:wx=>600, :boundingbox=>[154, -15, 575, 622]}, 49=>{:wx=>600, :boundingbox=>[98, 0, 515, 622]}, 50=>{:wx=>600, :boundingbox=>[70, 0, 568, 622]}, 51=>{:wx=>600, :boundingbox=>[82, -15, 538, 622]}, 52=>{:wx=>600, :boundingbox=>[108, 0, 541, 622]}, 53=>{:wx=>600, :boundingbox=>[99, -15, 589, 607]}, 54=>{:wx=>600, :boundingbox=>[155, -15, 629, 622]}, 55=>{:wx=>600, :boundingbox=>[182, 0, 612, 607]}, 56=>{:wx=>600, :boundingbox=>[132, -15, 588, 622]}, 57=>{:wx=>600, :boundingbox=>[93, -15, 574, 622]}, 58=>{:wx=>600, :boundingbox=>[238, -15, 441, 385]}, 59=>{:wx=>600, :boundingbox=>[157, -112, 441, 385]}, 60=>{:wx=>600, :boundingbox=>[96, 42, 610, 472]}, 61=>{:wx=>600, :boundingbox=>[109, 138, 600, 376]}, 62=>{:wx=>600, :boundingbox=>[85, 42, 599, 472]}, 63=>{:wx=>600, :boundingbox=>[222, -15, 583, 572]}, 64=>{:wx=>600, :boundingbox=>[127, -15, 582, 622]}, 65=>{:wx=>600, :boundingbox=>[3, 0, 607, 562]}, 66=>{:wx=>600, :boundingbox=>[43, 0, 616, 562]}, 67=>{:wx=>600, :boundingbox=>[93, -18, 655, 580]}, 68=>{:wx=>600, :boundingbox=>[43, 0, 645, 562]}, 69=>{:wx=>600, :boundingbox=>[53, 0, 660, 562]}, 70=>{:wx=>600, :boundingbox=>[53, 0, 660, 562]}, 71=>{:wx=>600, :boundingbox=>[83, -18, 645, 580]}, 72=>{:wx=>600, :boundingbox=>[32, 0, 687, 562]}, 73=>{:wx=>600, :boundingbox=>[96, 0, 623, 562]}, 74=>{:wx=>600, :boundingbox=>[52, -18, 685, 562]}, 75=>{:wx=>600, :boundingbox=>[38, 0, 671, 562]}, 76=>{:wx=>600, :boundingbox=>[47, 0, 607, 562]}, 77=>{:wx=>600, :boundingbox=>[4, 0, 715, 562]}, 78=>{:wx=>600, :boundingbox=>[7, -13, 712, 562]}, 79=>{:wx=>600, :boundingbox=>[94, -18, 625, 580]}, 80=>{:wx=>600, :boundingbox=>[79, 0, 644, 562]}, 81=>{:wx=>600, :boundingbox=>[95, -138, 625, 580]}, 82=>{:wx=>600, :boundingbox=>[38, 0, 598, 562]}, 83=>{:wx=>600, :boundingbox=>[76, -20, 650, 580]}, 84=>{:wx=>600, :boundingbox=>[108, 0, 665, 562]}, 85=>{:wx=>600, :boundingbox=>[125, -18, 702, 562]}, 86=>{:wx=>600, :boundingbox=>[105, -13, 723, 562]}, 87=>{:wx=>600, :boundingbox=>[106, -13, 722, 562]}, 88=>{:wx=>600, :boundingbox=>[23, 0, 675, 562]}, 89=>{:wx=>600, :boundingbox=>[133, 0, 695, 562]}, 90=>{:wx=>600, :boundingbox=>[86, 0, 610, 562]}, 91=>{:wx=>600, :boundingbox=>[246, -108, 574, 622]}, 92=>{:wx=>600, :boundingbox=>[249, -80, 468, 629]}, 93=>{:wx=>600, :boundingbox=>[135, -108, 463, 622]}, 94=>{:wx=>600, :boundingbox=>[175, 354, 587, 622]}, 95=>{:wx=>600, :boundingbox=>[-27, -125, 584, -75]}, 96=>{:wx=>600, :boundingbox=>[343, 328, 457, 562]}, 97=>{:wx=>600, :boundingbox=>[76, -15, 569, 441]}, 98=>{:wx=>600, :boundingbox=>[29, -15, 625, 629]}, 99=>{:wx=>600, :boundingbox=>[106, -15, 608, 441]}, 100=>{:wx=>600, :boundingbox=>[85, -15, 640, 629]}, 101=>{:wx=>600, :boundingbox=>[106, -15, 598, 441]}, 102=>{:wx=>600, :boundingbox=>[114, 0, 662, 629]}, 103=>{:wx=>600, :boundingbox=>[61, -157, 657, 441]}, 104=>{:wx=>600, :boundingbox=>[33, 0, 592, 629]}, 105=>{:wx=>600, :boundingbox=>[95, 0, 515, 657]}, 106=>{:wx=>600, :boundingbox=>[52, -157, 550, 657]}, 107=>{:wx=>600, :boundingbox=>[58, 0, 633, 629]}, 108=>{:wx=>600, :boundingbox=>[95, 0, 515, 629]}, 109=>{:wx=>600, :boundingbox=>[-5, 0, 615, 441]}, 110=>{:wx=>600, :boundingbox=>[26, 0, 585, 441]}, 111=>{:wx=>600, :boundingbox=>[102, -15, 588, 441]}, 112=>{:wx=>600, :boundingbox=>[-24, -157, 605, 441]}, 113=>{:wx=>600, :boundingbox=>[85, -157, 682, 441]}, 114=>{:wx=>600, :boundingbox=>[60, 0, 636, 441]}, 115=>{:wx=>600, :boundingbox=>[78, -15, 584, 441]}, 116=>{:wx=>600, :boundingbox=>[167, -15, 561, 561]}, 117=>{:wx=>600, :boundingbox=>[101, -15, 572, 426]}, 118=>{:wx=>600, :boundingbox=>[90, -10, 681, 426]}, 119=>{:wx=>600, :boundingbox=>[76, -10, 695, 426]}, 120=>{:wx=>600, :boundingbox=>[20, 0, 655, 426]}, 121=>{:wx=>600, :boundingbox=>[-4, -157, 683, 426]}, 122=>{:wx=>600, :boundingbox=>[99, 0, 593, 426]}, 123=>{:wx=>600, :boundingbox=>[233, -108, 569, 622]}, 124=>{:wx=>600, :boundingbox=>[222, -250, 485, 750]}, 125=>{:wx=>600, :boundingbox=>[140, -108, 477, 622]}, 126=>{:wx=>600, :boundingbox=>[116, 197, 600, 320]}, 161=>{:wx=>600, :boundingbox=>[225, -157, 445, 430]}, 162=>{:wx=>600, :boundingbox=>[151, -49, 588, 614]}, 163=>{:wx=>600, :boundingbox=>[124, -21, 621, 611]}, 164=>{:wx=>600, :boundingbox=>[84, -57, 646, 665]}, 165=>{:wx=>600, :boundingbox=>[120, 0, 693, 562]}, 166=>{:wx=>600, :boundingbox=>[-26, -143, 671, 622]}, 167=>{:wx=>600, :boundingbox=>[104, -78, 590, 580]}, 168=>{:wx=>600, :boundingbox=>[94, 58, 628, 506]}, 169=>{:wx=>600, :boundingbox=>[345, 328, 460, 562]}, 170=>{:wx=>600, :boundingbox=>[262, 328, 541, 562]}, 171=>{:wx=>600, :boundingbox=>[92, 70, 652, 446]}, 172=>{:wx=>600, :boundingbox=>[204, 70, 540, 446]}, 173=>{:wx=>600, :boundingbox=>[170, 70, 506, 446]}, 174=>{:wx=>600, :boundingbox=>[3, 0, 619, 629]}, 175=>{:wx=>600, :boundingbox=>[3, 0, 619, 629]}, 177=>{:wx=>600, :boundingbox=>[124, 231, 586, 285]}, 178=>{:wx=>600, :boundingbox=>[217, -78, 546, 580]}, 179=>{:wx=>600, :boundingbox=>[163, -78, 546, 580]}, 180=>{:wx=>600, :boundingbox=>[275, 189, 434, 327]}, 182=>{:wx=>600, :boundingbox=>[100, -78, 630, 562]}, 183=>{:wx=>600, :boundingbox=>[224, 130, 485, 383]}, 184=>{:wx=>600, :boundingbox=>[185, -134, 397, 100]}, 185=>{:wx=>600, :boundingbox=>[115, -134, 478, 100]}, 186=>{:wx=>600, :boundingbox=>[213, 328, 576, 562]}, 187=>{:wx=>600, :boundingbox=>[58, 70, 618, 446]}, 188=>{:wx=>600, :boundingbox=>[46, -15, 575, 111]}, 189=>{:wx=>600, :boundingbox=>[59, -15, 627, 622]}, 191=>{:wx=>600, :boundingbox=>[105, -157, 466, 430]}, 193=>{:wx=>600, :boundingbox=>[294, 497, 484, 672]}, 194=>{:wx=>600, :boundingbox=>[348, 497, 612, 672]}, 195=>{:wx=>600, :boundingbox=>[229, 477, 581, 654]}, 196=>{:wx=>600, :boundingbox=>[212, 489, 629, 606]}, 197=>{:wx=>600, :boundingbox=>[232, 525, 600, 565]}, 198=>{:wx=>600, :boundingbox=>[279, 501, 576, 609]}, 199=>{:wx=>600, :boundingbox=>[373, 537, 478, 640]}, 200=>{:wx=>600, :boundingbox=>[272, 537, 579, 640]}, 202=>{:wx=>600, :boundingbox=>[332, 463, 500, 627]}, 203=>{:wx=>600, :boundingbox=>[197, -151, 344, 10]}, 205=>{:wx=>600, :boundingbox=>[239, 497, 683, 672]}, 206=>{:wx=>600, :boundingbox=>[189, -172, 377, 4]}, 207=>{:wx=>600, :boundingbox=>[262, 492, 614, 669]}, 208=>{:wx=>600, :boundingbox=>[49, 231, 661, 285]}, 225=>{:wx=>600, :boundingbox=>[3, 0, 655, 562]}, 227=>{:wx=>600, :boundingbox=>[209, 249, 512, 580]}, 232=>{:wx=>600, :boundingbox=>[47, 0, 607, 562]}, 233=>{:wx=>600, :boundingbox=>[94, -80, 625, 629]}, 234=>{:wx=>600, :boundingbox=>[59, 0, 672, 562]}, 235=>{:wx=>600, :boundingbox=>[210, 249, 535, 580]}, 241=>{:wx=>600, :boundingbox=>[41, -15, 626, 441]}, 245=>{:wx=>600, :boundingbox=>[95, 0, 515, 426]}, 248=>{:wx=>600, :boundingbox=>[95, 0, 587, 629]}, 249=>{:wx=>600, :boundingbox=>[102, -80, 588, 506]}, 250=>{:wx=>600, :boundingbox=>[54, -15, 615, 441]}, 251=>{:wx=>600, :boundingbox=>[48, -15, 617, 629]}, -1=>{:wx=>600, :boundingbox=>[0, 0, 0, 0]}}
			courier_boldoblique_metrics = {32=>{:wx=>600, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>600, :boundingbox=>[215, -15, 495, 572]}, 34=>{:wx=>600, :boundingbox=>[211, 277, 585, 562]}, 35=>{:wx=>600, :boundingbox=>[88, -45, 641, 651]}, 36=>{:wx=>600, :boundingbox=>[87, -126, 630, 666]}, 37=>{:wx=>600, :boundingbox=>[101, -15, 625, 616]}, 38=>{:wx=>600, :boundingbox=>[61, -15, 595, 543]}, 39=>{:wx=>600, :boundingbox=>[229, 277, 543, 562]}, 40=>{:wx=>600, :boundingbox=>[265, -102, 592, 616]}, 41=>{:wx=>600, :boundingbox=>[117, -102, 444, 616]}, 42=>{:wx=>600, :boundingbox=>[179, 219, 598, 601]}, 43=>{:wx=>600, :boundingbox=>[114, 39, 596, 478]}, 44=>{:wx=>600, :boundingbox=>[99, -111, 430, 174]}, 45=>{:wx=>600, :boundingbox=>[143, 203, 567, 313]}, 46=>{:wx=>600, :boundingbox=>[206, -15, 427, 171]}, 47=>{:wx=>600, :boundingbox=>[90, -77, 626, 626]}, 48=>{:wx=>600, :boundingbox=>[135, -15, 593, 616]}, 49=>{:wx=>600, :boundingbox=>[93, 0, 562, 616]}, 50=>{:wx=>600, :boundingbox=>[61, 0, 594, 616]}, 51=>{:wx=>600, :boundingbox=>[71, -15, 571, 616]}, 52=>{:wx=>600, :boundingbox=>[81, 0, 559, 616]}, 53=>{:wx=>600, :boundingbox=>[77, -15, 621, 601]}, 54=>{:wx=>600, :boundingbox=>[135, -15, 652, 616]}, 55=>{:wx=>600, :boundingbox=>[147, 0, 622, 601]}, 56=>{:wx=>600, :boundingbox=>[115, -15, 604, 616]}, 57=>{:wx=>600, :boundingbox=>[75, -15, 592, 616]}, 58=>{:wx=>600, :boundingbox=>[205, -15, 480, 425]}, 59=>{:wx=>600, :boundingbox=>[99, -111, 481, 425]}, 60=>{:wx=>600, :boundingbox=>[120, 15, 613, 501]}, 61=>{:wx=>600, :boundingbox=>[96, 118, 614, 398]}, 62=>{:wx=>600, :boundingbox=>[97, 15, 589, 501]}, 63=>{:wx=>600, :boundingbox=>[183, -14, 592, 580]}, 64=>{:wx=>600, :boundingbox=>[65, -15, 642, 616]}, 65=>{:wx=>600, :boundingbox=>[-9, 0, 632, 562]}, 66=>{:wx=>600, :boundingbox=>[30, 0, 630, 562]}, 67=>{:wx=>600, :boundingbox=>[74, -18, 675, 580]}, 68=>{:wx=>600, :boundingbox=>[30, 0, 664, 562]}, 69=>{:wx=>600, :boundingbox=>[25, 0, 670, 562]}, 70=>{:wx=>600, :boundingbox=>[39, 0, 684, 562]}, 71=>{:wx=>600, :boundingbox=>[74, -18, 675, 580]}, 72=>{:wx=>600, :boundingbox=>[20, 0, 700, 562]}, 73=>{:wx=>600, :boundingbox=>[77, 0, 643, 562]}, 74=>{:wx=>600, :boundingbox=>[58, -18, 721, 562]}, 75=>{:wx=>600, :boundingbox=>[21, 0, 692, 562]}, 76=>{:wx=>600, :boundingbox=>[39, 0, 636, 562]}, 77=>{:wx=>600, :boundingbox=>[-2, 0, 722, 562]}, 78=>{:wx=>600, :boundingbox=>[8, -12, 730, 562]}, 79=>{:wx=>600, :boundingbox=>[74, -18, 645, 580]}, 80=>{:wx=>600, :boundingbox=>[48, 0, 643, 562]}, 81=>{:wx=>600, :boundingbox=>[83, -138, 636, 580]}, 82=>{:wx=>600, :boundingbox=>[24, 0, 617, 562]}, 83=>{:wx=>600, :boundingbox=>[54, -22, 673, 582]}, 84=>{:wx=>600, :boundingbox=>[86, 0, 679, 562]}, 85=>{:wx=>600, :boundingbox=>[101, -18, 716, 562]}, 86=>{:wx=>600, :boundingbox=>[84, 0, 733, 562]}, 87=>{:wx=>600, :boundingbox=>[79, 0, 738, 562]}, 88=>{:wx=>600, :boundingbox=>[12, 0, 690, 562]}, 89=>{:wx=>600, :boundingbox=>[109, 0, 709, 562]}, 90=>{:wx=>600, :boundingbox=>[62, 0, 637, 562]}, 91=>{:wx=>600, :boundingbox=>[223, -102, 606, 616]}, 92=>{:wx=>600, :boundingbox=>[222, -77, 496, 626]}, 93=>{:wx=>600, :boundingbox=>[103, -102, 486, 616]}, 94=>{:wx=>600, :boundingbox=>[171, 250, 556, 616]}, 95=>{:wx=>600, :boundingbox=>[-27, -125, 585, -75]}, 96=>{:wx=>600, :boundingbox=>[297, 277, 487, 562]}, 97=>{:wx=>600, :boundingbox=>[61, -15, 593, 454]}, 98=>{:wx=>600, :boundingbox=>[13, -15, 636, 626]}, 99=>{:wx=>600, :boundingbox=>[81, -15, 631, 459]}, 100=>{:wx=>600, :boundingbox=>[60, -15, 645, 626]}, 101=>{:wx=>600, :boundingbox=>[81, -15, 605, 454]}, 102=>{:wx=>600, :boundingbox=>[83, 0, 677, 626]}, 103=>{:wx=>600, :boundingbox=>[40, -146, 674, 454]}, 104=>{:wx=>600, :boundingbox=>[18, 0, 615, 626]}, 105=>{:wx=>600, :boundingbox=>[77, 0, 546, 658]}, 106=>{:wx=>600, :boundingbox=>[36, -146, 580, 658]}, 107=>{:wx=>600, :boundingbox=>[33, 0, 643, 626]}, 108=>{:wx=>600, :boundingbox=>[77, 0, 546, 626]}, 109=>{:wx=>600, :boundingbox=>[-22, 0, 649, 454]}, 110=>{:wx=>600, :boundingbox=>[18, 0, 615, 454]}, 111=>{:wx=>600, :boundingbox=>[71, -15, 622, 454]}, 112=>{:wx=>600, :boundingbox=>[-32, -142, 622, 454]}, 113=>{:wx=>600, :boundingbox=>[60, -142, 685, 454]}, 114=>{:wx=>600, :boundingbox=>[47, 0, 655, 454]}, 115=>{:wx=>600, :boundingbox=>[66, -17, 608, 459]}, 116=>{:wx=>600, :boundingbox=>[118, -15, 567, 562]}, 117=>{:wx=>600, :boundingbox=>[70, -15, 592, 439]}, 118=>{:wx=>600, :boundingbox=>[70, 0, 695, 439]}, 119=>{:wx=>600, :boundingbox=>[53, 0, 712, 439]}, 120=>{:wx=>600, :boundingbox=>[6, 0, 671, 439]}, 121=>{:wx=>600, :boundingbox=>[-21, -142, 695, 439]}, 122=>{:wx=>600, :boundingbox=>[81, 0, 614, 439]}, 123=>{:wx=>600, :boundingbox=>[203, -102, 595, 616]}, 124=>{:wx=>600, :boundingbox=>[201, -250, 505, 750]}, 125=>{:wx=>600, :boundingbox=>[114, -102, 506, 616]}, 126=>{:wx=>600, :boundingbox=>[120, 153, 590, 356]}, 161=>{:wx=>600, :boundingbox=>[196, -146, 477, 449]}, 162=>{:wx=>600, :boundingbox=>[121, -49, 605, 614]}, 163=>{:wx=>600, :boundingbox=>[106, -28, 650, 611]}, 164=>{:wx=>600, :boundingbox=>[22, -60, 708, 661]}, 165=>{:wx=>600, :boundingbox=>[98, 0, 710, 562]}, 166=>{:wx=>600, :boundingbox=>[-57, -131, 702, 616]}, 167=>{:wx=>600, :boundingbox=>[74, -70, 620, 580]}, 168=>{:wx=>600, :boundingbox=>[77, 49, 644, 517]}, 169=>{:wx=>600, :boundingbox=>[303, 277, 493, 562]}, 170=>{:wx=>600, :boundingbox=>[190, 277, 594, 562]}, 171=>{:wx=>600, :boundingbox=>[62, 70, 639, 446]}, 172=>{:wx=>600, :boundingbox=>[195, 70, 545, 446]}, 173=>{:wx=>600, :boundingbox=>[165, 70, 514, 446]}, 174=>{:wx=>600, :boundingbox=>[12, 0, 644, 626]}, 175=>{:wx=>600, :boundingbox=>[12, 0, 644, 626]}, 177=>{:wx=>600, :boundingbox=>[108, 203, 602, 313]}, 178=>{:wx=>600, :boundingbox=>[175, -70, 586, 580]}, 179=>{:wx=>600, :boundingbox=>[121, -70, 587, 580]}, 180=>{:wx=>600, :boundingbox=>[248, 165, 461, 351]}, 182=>{:wx=>600, :boundingbox=>[61, -70, 700, 580]}, 183=>{:wx=>600, :boundingbox=>[196, 132, 523, 430]}, 184=>{:wx=>600, :boundingbox=>[144, -142, 458, 143]}, 185=>{:wx=>600, :boundingbox=>[34, -142, 560, 143]}, 186=>{:wx=>600, :boundingbox=>[119, 277, 645, 562]}, 187=>{:wx=>600, :boundingbox=>[71, 70, 647, 446]}, 188=>{:wx=>600, :boundingbox=>[35, -15, 587, 116]}, 189=>{:wx=>600, :boundingbox=>[-45, -15, 743, 616]}, 191=>{:wx=>600, :boundingbox=>[100, -146, 509, 449]}, 193=>{:wx=>600, :boundingbox=>[272, 508, 503, 661]}, 194=>{:wx=>600, :boundingbox=>[312, 508, 609, 661]}, 195=>{:wx=>600, :boundingbox=>[212, 483, 607, 657]}, 196=>{:wx=>600, :boundingbox=>[199, 493, 643, 636]}, 197=>{:wx=>600, :boundingbox=>[195, 505, 637, 585]}, 198=>{:wx=>600, :boundingbox=>[217, 468, 652, 631]}, 199=>{:wx=>600, :boundingbox=>[348, 498, 493, 638]}, 200=>{:wx=>600, :boundingbox=>[246, 498, 595, 638]}, 202=>{:wx=>600, :boundingbox=>[319, 481, 528, 678]}, 203=>{:wx=>600, :boundingbox=>[168, -206, 368, 0]}, 205=>{:wx=>600, :boundingbox=>[171, 488, 729, 661]}, 206=>{:wx=>600, :boundingbox=>[143, -199, 367, 0]}, 207=>{:wx=>600, :boundingbox=>[238, 493, 633, 667]}, 208=>{:wx=>600, :boundingbox=>[33, 203, 677, 313]}, 225=>{:wx=>600, :boundingbox=>[-29, 0, 708, 562]}, 227=>{:wx=>600, :boundingbox=>[188, 196, 526, 580]}, 232=>{:wx=>600, :boundingbox=>[39, 0, 636, 562]}, 233=>{:wx=>600, :boundingbox=>[48, -22, 673, 584]}, 234=>{:wx=>600, :boundingbox=>[26, 0, 701, 562]}, 235=>{:wx=>600, :boundingbox=>[188, 196, 543, 580]}, 241=>{:wx=>600, :boundingbox=>[21, -15, 652, 454]}, 245=>{:wx=>600, :boundingbox=>[77, 0, 546, 439]}, 248=>{:wx=>600, :boundingbox=>[77, 0, 587, 626]}, 249=>{:wx=>600, :boundingbox=>[54, -24, 638, 463]}, 250=>{:wx=>600, :boundingbox=>[18, -15, 662, 454]}, 251=>{:wx=>600, :boundingbox=>[22, -15, 629, 626]}, -1=>{:wx=>600, :boundingbox=>[0, 0, 0, 0]}}
			symbol_metrics = {32=>{:wx=>250, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>333, :boundingbox=>[128, -17, 240, 672]}, 34=>{:wx=>713, :boundingbox=>[31, 0, 681, 705]}, 35=>{:wx=>500, :boundingbox=>[20, -16, 481, 673]}, 36=>{:wx=>549, :boundingbox=>[25, 0, 478, 707]}, 37=>{:wx=>833, :boundingbox=>[63, -36, 771, 655]}, 38=>{:wx=>778, :boundingbox=>[41, -18, 750, 661]}, 39=>{:wx=>439, :boundingbox=>[48, -17, 414, 500]}, 40=>{:wx=>333, :boundingbox=>[53, -191, 300, 673]}, 41=>{:wx=>333, :boundingbox=>[30, -191, 277, 673]}, 42=>{:wx=>500, :boundingbox=>[65, 134, 427, 551]}, 43=>{:wx=>549, :boundingbox=>[10, 0, 539, 533]}, 44=>{:wx=>250, :boundingbox=>[56, -152, 194, 104]}, 45=>{:wx=>549, :boundingbox=>[11, 233, 535, 288]}, 46=>{:wx=>250, :boundingbox=>[69, -17, 181, 95]}, 47=>{:wx=>278, :boundingbox=>[0, -18, 254, 646]}, 48=>{:wx=>500, :boundingbox=>[24, -14, 476, 685]}, 49=>{:wx=>500, :boundingbox=>[117, 0, 390, 673]}, 50=>{:wx=>500, :boundingbox=>[25, 0, 475, 685]}, 51=>{:wx=>500, :boundingbox=>[43, -14, 435, 685]}, 52=>{:wx=>500, :boundingbox=>[15, 0, 469, 685]}, 53=>{:wx=>500, :boundingbox=>[32, -14, 445, 690]}, 54=>{:wx=>500, :boundingbox=>[34, -14, 468, 685]}, 55=>{:wx=>500, :boundingbox=>[24, -16, 448, 673]}, 56=>{:wx=>500, :boundingbox=>[56, -14, 445, 685]}, 57=>{:wx=>500, :boundingbox=>[30, -18, 459, 685]}, 58=>{:wx=>278, :boundingbox=>[81, -17, 193, 460]}, 59=>{:wx=>278, :boundingbox=>[83, -152, 221, 460]}, 60=>{:wx=>549, :boundingbox=>[26, 0, 523, 522]}, 61=>{:wx=>549, :boundingbox=>[11, 141, 537, 390]}, 62=>{:wx=>549, :boundingbox=>[26, 0, 523, 522]}, 63=>{:wx=>444, :boundingbox=>[70, -17, 412, 686]}, 64=>{:wx=>549, :boundingbox=>[11, 0, 537, 475]}, 65=>{:wx=>722, :boundingbox=>[4, 0, 684, 673]}, 66=>{:wx=>667, :boundingbox=>[29, 0, 592, 673]}, 67=>{:wx=>722, :boundingbox=>[-9, 0, 704, 673]}, 68=>{:wx=>612, :boundingbox=>[6, 0, 608, 688]}, 69=>{:wx=>611, :boundingbox=>[32, 0, 617, 673]}, 70=>{:wx=>763, :boundingbox=>[26, 0, 741, 673]}, 71=>{:wx=>603, :boundingbox=>[24, 0, 609, 673]}, 72=>{:wx=>722, :boundingbox=>[39, 0, 729, 673]}, 73=>{:wx=>333, :boundingbox=>[32, 0, 316, 673]}, 74=>{:wx=>631, :boundingbox=>[18, -18, 623, 689]}, 75=>{:wx=>722, :boundingbox=>[35, 0, 722, 673]}, 76=>{:wx=>686, :boundingbox=>[6, 0, 680, 688]}, 77=>{:wx=>889, :boundingbox=>[28, 0, 887, 673]}, 78=>{:wx=>722, :boundingbox=>[29, -8, 720, 673]}, 79=>{:wx=>722, :boundingbox=>[41, -17, 715, 685]}, 80=>{:wx=>768, :boundingbox=>[25, 0, 745, 673]}, 81=>{:wx=>741, :boundingbox=>[41, -17, 715, 685]}, 82=>{:wx=>556, :boundingbox=>[28, 0, 563, 673]}, 83=>{:wx=>592, :boundingbox=>[5, 0, 589, 673]}, 84=>{:wx=>611, :boundingbox=>[33, 0, 607, 673]}, 85=>{:wx=>690, :boundingbox=>[-8, 0, 694, 673]}, 86=>{:wx=>439, :boundingbox=>[40, -233, 436, 500]}, 87=>{:wx=>768, :boundingbox=>[34, 0, 736, 688]}, 88=>{:wx=>645, :boundingbox=>[40, 0, 599, 673]}, 89=>{:wx=>795, :boundingbox=>[15, 0, 781, 684]}, 90=>{:wx=>611, :boundingbox=>[44, 0, 636, 673]}, 91=>{:wx=>333, :boundingbox=>[86, -155, 299, 674]}, 92=>{:wx=>863, :boundingbox=>[163, 0, 701, 487]}, 93=>{:wx=>333, :boundingbox=>[33, -155, 246, 674]}, 94=>{:wx=>658, :boundingbox=>[15, 0, 652, 674]}, 95=>{:wx=>500, :boundingbox=>[-2, -125, 502, -75]}, 96=>{:wx=>500, :boundingbox=>[480, 881, 1090, 917]}, 97=>{:wx=>631, :boundingbox=>[41, -18, 622, 500]}, 98=>{:wx=>549, :boundingbox=>[61, -223, 515, 741]}, 99=>{:wx=>549, :boundingbox=>[12, -231, 522, 499]}, 100=>{:wx=>494, :boundingbox=>[40, -19, 481, 740]}, 101=>{:wx=>439, :boundingbox=>[22, -19, 427, 502]}, 102=>{:wx=>521, :boundingbox=>[28, -224, 492, 673]}, 103=>{:wx=>411, :boundingbox=>[5, -225, 484, 499]}, 104=>{:wx=>603, :boundingbox=>[0, -202, 527, 514]}, 105=>{:wx=>329, :boundingbox=>[0, -17, 301, 503]}, 106=>{:wx=>603, :boundingbox=>[36, -224, 587, 499]}, 107=>{:wx=>549, :boundingbox=>[33, 0, 558, 501]}, 108=>{:wx=>549, :boundingbox=>[24, -17, 548, 739]}, 109=>{:wx=>576, :boundingbox=>[33, -223, 567, 500]}, 110=>{:wx=>521, :boundingbox=>[-9, -16, 475, 507]}, 111=>{:wx=>549, :boundingbox=>[35, -19, 501, 499]}, 112=>{:wx=>549, :boundingbox=>[10, -19, 530, 487]}, 113=>{:wx=>521, :boundingbox=>[43, -17, 485, 690]}, 114=>{:wx=>549, :boundingbox=>[50, -230, 490, 499]}, 115=>{:wx=>603, :boundingbox=>[30, -21, 588, 500]}, 116=>{:wx=>439, :boundingbox=>[10, -19, 418, 500]}, 117=>{:wx=>576, :boundingbox=>[7, -18, 535, 507]}, 118=>{:wx=>713, :boundingbox=>[12, -18, 671, 583]}, 119=>{:wx=>686, :boundingbox=>[42, -17, 684, 500]}, 120=>{:wx=>493, :boundingbox=>[27, -224, 469, 766]}, 121=>{:wx=>686, :boundingbox=>[12, -228, 701, 500]}, 122=>{:wx=>494, :boundingbox=>[60, -225, 467, 756]}, 123=>{:wx=>480, :boundingbox=>[58, -183, 397, 673]}, 124=>{:wx=>200, :boundingbox=>[65, -293, 135, 707]}, 125=>{:wx=>480, :boundingbox=>[79, -183, 418, 673]}, 126=>{:wx=>549, :boundingbox=>[17, 203, 529, 307]}, 160=>{:wx=>750, :boundingbox=>[20, -12, 714, 685]}, 161=>{:wx=>620, :boundingbox=>[-2, 0, 610, 685]}, 162=>{:wx=>247, :boundingbox=>[27, 459, 228, 735]}, 163=>{:wx=>549, :boundingbox=>[29, 0, 526, 639]}, 164=>{:wx=>167, :boundingbox=>[-180, -12, 340, 677]}, 165=>{:wx=>713, :boundingbox=>[26, 124, 688, 404]}, 166=>{:wx=>500, :boundingbox=>[2, -193, 494, 686]}, 167=>{:wx=>753, :boundingbox=>[86, -26, 660, 533]}, 168=>{:wx=>753, :boundingbox=>[142, -36, 600, 550]}, 169=>{:wx=>753, :boundingbox=>[117, -33, 631, 532]}, 170=>{:wx=>753, :boundingbox=>[113, -36, 629, 548]}, 171=>{:wx=>1042, :boundingbox=>[24, -15, 1024, 511]}, 172=>{:wx=>987, :boundingbox=>[32, -15, 942, 511]}, 173=>{:wx=>603, :boundingbox=>[45, 0, 571, 910]}, 174=>{:wx=>987, :boundingbox=>[49, -15, 959, 511]}, 175=>{:wx=>603, :boundingbox=>[45, -22, 571, 888]}, 176=>{:wx=>400, :boundingbox=>[50, 385, 350, 685]}, 177=>{:wx=>549, :boundingbox=>[10, 0, 539, 645]}, 178=>{:wx=>411, :boundingbox=>[20, 459, 413, 737]}, 179=>{:wx=>549, :boundingbox=>[29, 0, 526, 639]}, 180=>{:wx=>549, :boundingbox=>[17, 8, 533, 524]}, 181=>{:wx=>713, :boundingbox=>[27, 123, 639, 404]}, 182=>{:wx=>494, :boundingbox=>[26, -20, 462, 746]}, 183=>{:wx=>460, :boundingbox=>[50, 113, 410, 473]}, 184=>{:wx=>549, :boundingbox=>[10, 71, 536, 456]}, 185=>{:wx=>549, :boundingbox=>[15, -25, 540, 549]}, 186=>{:wx=>549, :boundingbox=>[14, 82, 538, 443]}, 187=>{:wx=>549, :boundingbox=>[14, 135, 527, 394]}, 188=>{:wx=>1000, :boundingbox=>[111, -17, 889, 95]}, 189=>{:wx=>603, :boundingbox=>[280, -120, 336, 1010]}, 190=>{:wx=>1000, :boundingbox=>[-60, 220, 1050, 276]}, 191=>{:wx=>658, :boundingbox=>[15, -16, 602, 629]}, 192=>{:wx=>823, :boundingbox=>[175, -18, 661, 658]}, 193=>{:wx=>686, :boundingbox=>[10, -53, 578, 740]}, 194=>{:wx=>795, :boundingbox=>[26, -15, 759, 734]}, 195=>{:wx=>987, :boundingbox=>[159, -211, 870, 573]}, 196=>{:wx=>768, :boundingbox=>[43, -17, 733, 673]}, 197=>{:wx=>768, :boundingbox=>[43, -15, 733, 675]}, 198=>{:wx=>823, :boundingbox=>[39, -24, 781, 719]}, 199=>{:wx=>768, :boundingbox=>[40, 0, 732, 509]}, 200=>{:wx=>768, :boundingbox=>[40, -17, 732, 492]}, 201=>{:wx=>713, :boundingbox=>[20, 0, 673, 470]}, 202=>{:wx=>713, :boundingbox=>[20, -125, 673, 470]}, 203=>{:wx=>713, :boundingbox=>[36, -70, 690, 540]}, 204=>{:wx=>713, :boundingbox=>[37, 0, 690, 470]}, 205=>{:wx=>713, :boundingbox=>[37, -125, 690, 470]}, 206=>{:wx=>713, :boundingbox=>[45, 0, 505, 468]}, 207=>{:wx=>713, :boundingbox=>[45, -58, 505, 555]}, 208=>{:wx=>768, :boundingbox=>[26, 0, 738, 673]}, 209=>{:wx=>713, :boundingbox=>[36, -19, 681, 718]}, 210=>{:wx=>790, :boundingbox=>[50, -17, 740, 673]}, 211=>{:wx=>790, :boundingbox=>[51, -15, 741, 675]}, 212=>{:wx=>890, :boundingbox=>[18, 293, 855, 673]}, 213=>{:wx=>823, :boundingbox=>[25, -101, 803, 751]}, 214=>{:wx=>549, :boundingbox=>[10, -38, 515, 917]}, 215=>{:wx=>250, :boundingbox=>[69, 210, 169, 310]}, 216=>{:wx=>713, :boundingbox=>[15, 0, 680, 288]}, 217=>{:wx=>603, :boundingbox=>[23, 0, 583, 454]}, 218=>{:wx=>603, :boundingbox=>[30, 0, 578, 477]}, 219=>{:wx=>1042, :boundingbox=>[27, -20, 1023, 510]}, 220=>{:wx=>987, :boundingbox=>[30, -15, 939, 513]}, 221=>{:wx=>603, :boundingbox=>[39, 2, 567, 911]}, 222=>{:wx=>987, :boundingbox=>[45, -20, 954, 508]}, 223=>{:wx=>603, :boundingbox=>[44, -19, 572, 890]}, 224=>{:wx=>494, :boundingbox=>[18, 0, 466, 745]}, 225=>{:wx=>329, :boundingbox=>[25, -198, 306, 746]}, 226=>{:wx=>790, :boundingbox=>[50, -20, 740, 670]}, 227=>{:wx=>790, :boundingbox=>[49, -15, 739, 675]}, 228=>{:wx=>786, :boundingbox=>[5, 293, 725, 673]}, 229=>{:wx=>713, :boundingbox=>[14, -108, 695, 752]}, 230=>{:wx=>384, :boundingbox=>[24, -293, 436, 926]}, 231=>{:wx=>384, :boundingbox=>[24, -85, 108, 925]}, 232=>{:wx=>384, :boundingbox=>[24, -293, 436, 926]}, 233=>{:wx=>384, :boundingbox=>[0, -80, 349, 926]}, 234=>{:wx=>384, :boundingbox=>[0, -79, 77, 925]}, 235=>{:wx=>384, :boundingbox=>[0, -80, 349, 926]}, 236=>{:wx=>494, :boundingbox=>[209, -85, 445, 925]}, 237=>{:wx=>494, :boundingbox=>[20, -85, 284, 935]}, 238=>{:wx=>494, :boundingbox=>[209, -75, 445, 935]}, 239=>{:wx=>494, :boundingbox=>[209, -85, 284, 935]}, 241=>{:wx=>329, :boundingbox=>[21, -198, 302, 746]}, 242=>{:wx=>274, :boundingbox=>[2, -107, 291, 916]}, 243=>{:wx=>686, :boundingbox=>[308, -88, 675, 920]}, 244=>{:wx=>686, :boundingbox=>[308, -88, 378, 975]}, 245=>{:wx=>686, :boundingbox=>[11, -87, 378, 921]}, 246=>{:wx=>384, :boundingbox=>[54, -293, 466, 926]}, 247=>{:wx=>384, :boundingbox=>[382, -85, 466, 925]}, 248=>{:wx=>384, :boundingbox=>[54, -293, 466, 926]}, 249=>{:wx=>384, :boundingbox=>[22, -80, 371, 926]}, 250=>{:wx=>384, :boundingbox=>[294, -79, 371, 925]}, 251=>{:wx=>384, :boundingbox=>[22, -80, 371, 926]}, 252=>{:wx=>494, :boundingbox=>[48, -85, 284, 925]}, 253=>{:wx=>494, :boundingbox=>[209, -85, 473, 935]}, 254=>{:wx=>494, :boundingbox=>[48, -75, 284, 935]}, -1=>{:wx=>790, :boundingbox=>[56, -3, 733, 808]}}
			zapfdingbats_metrics = {32=>{:wx=>278, :boundingbox=>[0, 0, 0, 0]}, 33=>{:wx=>974, :boundingbox=>[35, 72, 939, 621]}, 34=>{:wx=>961, :boundingbox=>[35, 81, 927, 611]}, 35=>{:wx=>974, :boundingbox=>[35, 72, 939, 621]}, 36=>{:wx=>980, :boundingbox=>[35, 0, 945, 692]}, 37=>{:wx=>719, :boundingbox=>[34, 139, 685, 566]}, 38=>{:wx=>789, :boundingbox=>[35, -14, 755, 705]}, 39=>{:wx=>790, :boundingbox=>[35, -14, 755, 705]}, 40=>{:wx=>791, :boundingbox=>[35, -13, 761, 705]}, 41=>{:wx=>690, :boundingbox=>[34, 138, 655, 553]}, 42=>{:wx=>960, :boundingbox=>[35, 123, 925, 568]}, 43=>{:wx=>939, :boundingbox=>[35, 134, 904, 559]}, 44=>{:wx=>549, :boundingbox=>[29, -11, 516, 705]}, 45=>{:wx=>855, :boundingbox=>[34, 59, 820, 632]}, 46=>{:wx=>911, :boundingbox=>[35, 50, 876, 642]}, 47=>{:wx=>933, :boundingbox=>[35, 139, 899, 550]}, 48=>{:wx=>911, :boundingbox=>[35, 50, 876, 642]}, 49=>{:wx=>945, :boundingbox=>[35, 139, 909, 553]}, 50=>{:wx=>974, :boundingbox=>[35, 104, 938, 587]}, 51=>{:wx=>755, :boundingbox=>[34, -13, 721, 705]}, 52=>{:wx=>846, :boundingbox=>[36, -14, 811, 705]}, 53=>{:wx=>762, :boundingbox=>[35, 0, 727, 692]}, 54=>{:wx=>761, :boundingbox=>[35, 0, 727, 692]}, 55=>{:wx=>571, :boundingbox=>[-1, -68, 571, 661]}, 56=>{:wx=>677, :boundingbox=>[36, -13, 642, 705]}, 57=>{:wx=>763, :boundingbox=>[35, 0, 728, 692]}, 58=>{:wx=>760, :boundingbox=>[35, 0, 726, 692]}, 59=>{:wx=>759, :boundingbox=>[35, 0, 725, 692]}, 60=>{:wx=>754, :boundingbox=>[35, 0, 720, 692]}, 61=>{:wx=>494, :boundingbox=>[35, 0, 460, 692]}, 62=>{:wx=>552, :boundingbox=>[35, 0, 517, 692]}, 63=>{:wx=>537, :boundingbox=>[35, 0, 503, 692]}, 64=>{:wx=>577, :boundingbox=>[35, 96, 542, 596]}, 65=>{:wx=>692, :boundingbox=>[35, -14, 657, 705]}, 66=>{:wx=>786, :boundingbox=>[35, -14, 751, 705]}, 67=>{:wx=>788, :boundingbox=>[35, -14, 752, 705]}, 68=>{:wx=>788, :boundingbox=>[35, -14, 753, 705]}, 69=>{:wx=>790, :boundingbox=>[35, -14, 756, 705]}, 70=>{:wx=>793, :boundingbox=>[35, -13, 759, 705]}, 71=>{:wx=>794, :boundingbox=>[35, -13, 759, 705]}, 72=>{:wx=>816, :boundingbox=>[35, -14, 782, 705]}, 73=>{:wx=>823, :boundingbox=>[35, -14, 787, 705]}, 74=>{:wx=>789, :boundingbox=>[35, -14, 754, 705]}, 75=>{:wx=>841, :boundingbox=>[35, -14, 807, 705]}, 76=>{:wx=>823, :boundingbox=>[35, -14, 789, 705]}, 77=>{:wx=>833, :boundingbox=>[35, -14, 798, 705]}, 78=>{:wx=>816, :boundingbox=>[35, -13, 782, 705]}, 79=>{:wx=>831, :boundingbox=>[35, -14, 796, 705]}, 80=>{:wx=>923, :boundingbox=>[35, -14, 888, 705]}, 81=>{:wx=>744, :boundingbox=>[35, 0, 710, 692]}, 82=>{:wx=>723, :boundingbox=>[35, 0, 688, 692]}, 83=>{:wx=>749, :boundingbox=>[35, 0, 714, 692]}, 84=>{:wx=>790, :boundingbox=>[34, -14, 756, 705]}, 85=>{:wx=>792, :boundingbox=>[35, -14, 758, 705]}, 86=>{:wx=>695, :boundingbox=>[35, -14, 661, 706]}, 87=>{:wx=>776, :boundingbox=>[35, -6, 741, 699]}, 88=>{:wx=>768, :boundingbox=>[35, -7, 734, 699]}, 89=>{:wx=>792, :boundingbox=>[35, -14, 757, 705]}, 90=>{:wx=>759, :boundingbox=>[35, 0, 725, 692]}, 91=>{:wx=>707, :boundingbox=>[35, -13, 672, 704]}, 92=>{:wx=>708, :boundingbox=>[35, -14, 672, 705]}, 93=>{:wx=>682, :boundingbox=>[35, -14, 647, 705]}, 94=>{:wx=>701, :boundingbox=>[35, -14, 666, 705]}, 95=>{:wx=>826, :boundingbox=>[35, -14, 791, 705]}, 96=>{:wx=>815, :boundingbox=>[35, -14, 780, 705]}, 97=>{:wx=>789, :boundingbox=>[35, -14, 754, 705]}, 98=>{:wx=>789, :boundingbox=>[35, -14, 754, 705]}, 99=>{:wx=>707, :boundingbox=>[34, -14, 673, 705]}, 100=>{:wx=>687, :boundingbox=>[36, 0, 651, 692]}, 101=>{:wx=>696, :boundingbox=>[35, 0, 661, 691]}, 102=>{:wx=>689, :boundingbox=>[35, 0, 655, 692]}, 103=>{:wx=>786, :boundingbox=>[34, -14, 751, 705]}, 104=>{:wx=>787, :boundingbox=>[35, -14, 752, 705]}, 105=>{:wx=>713, :boundingbox=>[35, -14, 678, 705]}, 106=>{:wx=>791, :boundingbox=>[35, -14, 756, 705]}, 107=>{:wx=>785, :boundingbox=>[36, -14, 751, 705]}, 108=>{:wx=>791, :boundingbox=>[35, -14, 757, 705]}, 109=>{:wx=>873, :boundingbox=>[35, -14, 838, 705]}, 110=>{:wx=>761, :boundingbox=>[35, 0, 726, 692]}, 111=>{:wx=>762, :boundingbox=>[35, 0, 727, 692]}, 112=>{:wx=>762, :boundingbox=>[35, 0, 727, 692]}, 113=>{:wx=>759, :boundingbox=>[35, 0, 725, 692]}, 114=>{:wx=>759, :boundingbox=>[35, 0, 725, 692]}, 115=>{:wx=>892, :boundingbox=>[35, 0, 858, 705]}, 116=>{:wx=>892, :boundingbox=>[35, -14, 858, 692]}, 117=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 118=>{:wx=>784, :boundingbox=>[35, -14, 749, 705]}, 119=>{:wx=>438, :boundingbox=>[35, -14, 403, 705]}, 120=>{:wx=>138, :boundingbox=>[35, 0, 104, 692]}, 121=>{:wx=>277, :boundingbox=>[35, 0, 242, 692]}, 122=>{:wx=>415, :boundingbox=>[35, 0, 380, 692]}, 123=>{:wx=>392, :boundingbox=>[35, 263, 357, 705]}, 124=>{:wx=>392, :boundingbox=>[34, 263, 357, 705]}, 125=>{:wx=>668, :boundingbox=>[35, 263, 633, 705]}, 126=>{:wx=>668, :boundingbox=>[36, 263, 634, 705]}, 128=>{:wx=>390, :boundingbox=>[35, -14, 356, 705]}, 129=>{:wx=>390, :boundingbox=>[35, -14, 355, 705]}, 130=>{:wx=>317, :boundingbox=>[35, 0, 283, 692]}, 131=>{:wx=>317, :boundingbox=>[35, 0, 283, 692]}, 132=>{:wx=>276, :boundingbox=>[35, 0, 242, 692]}, 133=>{:wx=>276, :boundingbox=>[35, 0, 242, 692]}, 134=>{:wx=>509, :boundingbox=>[35, 0, 475, 692]}, 135=>{:wx=>509, :boundingbox=>[35, 0, 475, 692]}, 136=>{:wx=>410, :boundingbox=>[35, 0, 375, 692]}, 137=>{:wx=>410, :boundingbox=>[35, 0, 375, 692]}, 138=>{:wx=>234, :boundingbox=>[35, -14, 199, 705]}, 139=>{:wx=>234, :boundingbox=>[35, -14, 199, 705]}, 140=>{:wx=>334, :boundingbox=>[35, 0, 299, 692]}, 141=>{:wx=>334, :boundingbox=>[35, 0, 299, 692]}, 161=>{:wx=>732, :boundingbox=>[35, -143, 697, 806]}, 162=>{:wx=>544, :boundingbox=>[56, -14, 488, 706]}, 163=>{:wx=>544, :boundingbox=>[34, -14, 508, 705]}, 164=>{:wx=>910, :boundingbox=>[35, 40, 875, 651]}, 165=>{:wx=>667, :boundingbox=>[35, -14, 633, 705]}, 166=>{:wx=>760, :boundingbox=>[35, -14, 726, 705]}, 167=>{:wx=>760, :boundingbox=>[0, 121, 758, 569]}, 168=>{:wx=>776, :boundingbox=>[35, 0, 741, 705]}, 169=>{:wx=>595, :boundingbox=>[34, -14, 560, 705]}, 170=>{:wx=>694, :boundingbox=>[35, -14, 659, 705]}, 171=>{:wx=>626, :boundingbox=>[34, 0, 591, 705]}, 172=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 173=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 174=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 175=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 176=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 177=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 178=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 179=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 180=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 181=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 182=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 183=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 184=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 185=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 186=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 187=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 188=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 189=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 190=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 191=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 192=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 193=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 194=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 195=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 196=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 197=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 198=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 199=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 200=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 201=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 202=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 203=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 204=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 205=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 206=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 207=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 208=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 209=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 210=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 211=>{:wx=>788, :boundingbox=>[35, -14, 754, 705]}, 212=>{:wx=>894, :boundingbox=>[35, 58, 860, 634]}, 213=>{:wx=>838, :boundingbox=>[35, 152, 803, 540]}, 214=>{:wx=>1016, :boundingbox=>[34, 152, 981, 540]}, 215=>{:wx=>458, :boundingbox=>[35, -127, 422, 820]}, 216=>{:wx=>748, :boundingbox=>[35, 94, 698, 597]}, 217=>{:wx=>924, :boundingbox=>[35, 140, 890, 552]}, 218=>{:wx=>748, :boundingbox=>[35, 94, 698, 597]}, 219=>{:wx=>918, :boundingbox=>[35, 166, 884, 526]}, 220=>{:wx=>927, :boundingbox=>[35, 32, 892, 660]}, 221=>{:wx=>928, :boundingbox=>[35, 129, 891, 562]}, 222=>{:wx=>928, :boundingbox=>[35, 128, 893, 563]}, 223=>{:wx=>834, :boundingbox=>[35, 155, 799, 537]}, 224=>{:wx=>873, :boundingbox=>[35, 93, 838, 599]}, 225=>{:wx=>828, :boundingbox=>[35, 104, 791, 588]}, 226=>{:wx=>924, :boundingbox=>[35, 98, 889, 594]}, 227=>{:wx=>924, :boundingbox=>[35, 98, 889, 594]}, 228=>{:wx=>917, :boundingbox=>[35, 0, 882, 692]}, 229=>{:wx=>930, :boundingbox=>[35, 84, 896, 608]}, 230=>{:wx=>931, :boundingbox=>[35, 84, 896, 608]}, 231=>{:wx=>463, :boundingbox=>[35, -99, 429, 791]}, 232=>{:wx=>883, :boundingbox=>[35, 71, 848, 623]}, 233=>{:wx=>836, :boundingbox=>[35, 44, 802, 648]}, 234=>{:wx=>836, :boundingbox=>[35, 44, 802, 648]}, 235=>{:wx=>867, :boundingbox=>[35, 101, 832, 591]}, 236=>{:wx=>867, :boundingbox=>[35, 101, 832, 591]}, 237=>{:wx=>696, :boundingbox=>[35, 44, 661, 648]}, 238=>{:wx=>696, :boundingbox=>[35, 44, 661, 648]}, 239=>{:wx=>874, :boundingbox=>[35, 77, 840, 619]}, 241=>{:wx=>874, :boundingbox=>[35, 73, 840, 615]}, 242=>{:wx=>760, :boundingbox=>[35, 0, 725, 692]}, 243=>{:wx=>946, :boundingbox=>[35, 160, 911, 533]}, 244=>{:wx=>771, :boundingbox=>[34, 37, 736, 655]}, 245=>{:wx=>865, :boundingbox=>[35, 207, 830, 481]}, 246=>{:wx=>771, :boundingbox=>[34, 37, 736, 655]}, 247=>{:wx=>888, :boundingbox=>[34, -19, 853, 712]}, 248=>{:wx=>967, :boundingbox=>[35, 124, 932, 568]}, 249=>{:wx=>888, :boundingbox=>[34, -19, 853, 712]}, 250=>{:wx=>831, :boundingbox=>[35, 113, 796, 579]}, 251=>{:wx=>873, :boundingbox=>[36, 118, 838, 578]}, 252=>{:wx=>927, :boundingbox=>[35, 150, 891, 542]}, 253=>{:wx=>970, :boundingbox=>[35, 76, 931, 616]}, 254=>{:wx=>918, :boundingbox=>[34, 99, 884, 593]}}

			# make two correlating arrays (indexes reffer to the same data), one for font names and the other for the fonts matrics.
			fonts_metrics_array = [	times_metrics, times_bold_metrics, times_italic_metrics, times_bolditalic_metrics,
									helvetica_metrics, helvetica_bold_metrics, helvetica_oblique_metrics, helvetica_oblique_metrics,
									courier_metrics, courier_bold_metrics, courier_oblique_metrics, courier_boldoblique_metrics,
									symbol_metrics, zapfdingbats_metrics]
			fonts_names_array = [:"Times-Roman",
			:"Times-Bold",
			:"Times-Italic",
			:"Times-BoldItalic",
			:Helvetica,
			:"Helvetica-Bold",
			:"Helvetica-BoldOblique",
			:"Helvetica-Oblique",
			:Courier,
			:"Courier-Bold",
			:"Courier-Oblique",
			:"Courier-BoldOblique",
			:Symbol,
			:ZapfDingbats ]

			# create the font object and register the font for each one of the 14 fonts
			fonts_names_array.each_index do |i|
				CombinePDF::Fonts.register_font fonts_names_array[i], fonts_metrics_array[i], { Type: :Font, Subtype: :Type1, BaseFont: fonts_names_array[i]}
			end
		end

		protected
		# the Hash listing all the fonts.
		FONTS_LIBRARY = {}
	end
end






